# -*- coding: utf-8 -*-
"""Reactions."""

from ._reactions import CustomReaction, Limit, LimitChildren, UniqueAttributes, reaction

__all__ = [
    "reaction",
    "CustomReaction",
    "UniqueAttributes",
    "LimitChildren",
    "Limit",
]



# -*- coding: utf-8 -*-
"""Exceptions."""

from ._applications import ActionObserversFailedError, RejectChangeException
from ._history import HistoryError
from ._structures import SerializationError

__all__ = [
    "ActionObserversFailedError",
    "RejectChangeException",
    "HistoryError",
    "SerializationError",
]



# -*- coding: utf-8 -*-

import pytest

from objetto.utils.caller_module import get_caller_module


def test_get_caller_module():
    def func():
        return get_caller_module()

    assert func() == __name__


if __name__ == "__main__":
    pytest.main()



# -*- coding: utf-8 -*-
"""Reactions."""

from collections import Counter as ValueCounter
from collections import defaultdict
from typing import TYPE_CHECKING, cast

from six import integer_types, iteritems, string_types

from ._applications import Phase, RejectChangeException
from ._bases import MISSING, final
from ._changes import BaseAtomicChange, Update
from ._data import InteractiveDictData
from ._objects import UNIQUE_ATTRIBUTES_METADATA_KEY, BaseReaction, Object
from ._states import DictState
from .utils.reraise_context import ReraiseContext
from .utils.type_checking import assert_is_callable, assert_is_instance

if TYPE_CHECKING:
    from typing import Any, Callable, Counter, Dict, FrozenSet, Mapping, Optional, Union

    from ._applications import Action
    from ._objects import BaseObject

    ReactionDecorator = Callable[
        [Callable[[BaseObject, Action, Phase], None]], "CustomReaction"
    ]

__all__ = [
    "reaction",
    "CustomReaction",
    "UniqueAttributes",
    "LimitChildren",
    "Limit",
]


# noinspection PyAbstractClass
def reaction(
    func=None,  # type: Optional[Callable[[BaseObject, Action, Phase], None]]
    priority=None,  # type: Optional[int]
):
    # type: (...) -> Union[CustomReaction, ReactionDecorator]
    """
    Decorates an object's method into a custom reaction.
    Reaction methods are called automatically when an action propagates up the
    hierarchy during the 'PRE' and 'POST' phases.

    . code:: python

        >>> from objetto.applications import Application
        >>> from objetto.objects import Object, attribute
        >>> from objetto.reactions import reaction

        >>> class MyObject(Object):
        ...     value = attribute(int, default=0)
        ...
        ...     @reaction
        ...     def __on_received(self, action, phase):
        ...         if not self._initializing:
        ...             print(("LAST -", action.change.name, phase))
        ...
        ...     @reaction(priority=1)
        ...     def __on_received_first(self, action, phase):
        ...         if not self._initializing:
        ...             print(("FIRST -", action.change.name, phase))
        ...
        >>> app = Application()
        >>> my_obj = MyObject(app)
        >>> my_obj.value = 42
        ('FIRST -', 'Update Attributes', <Phase.PRE: 'PRE'>)
        ('LAST -', 'Update Attributes', <Phase.PRE: 'PRE'>)
        ('FIRST -', 'Update Attributes', <Phase.POST: 'POST'>)
        ('LAST -', 'Update Attributes', <Phase.POST: 'POST'>)

    :param func: Method to be decorated or `None`.
    :type func: function

    :param priority: Priority.
    :type priority: int or None

    :return: Decorated custom reaction method or decorator.
    :rtype: objetto.reactions.CustomReaction
    """

    def _reaction(func_):
        # type: (Callable[[BaseObject, Action, Phase], None]) -> CustomReaction
        """
        Reaction method decorator.

        :param func_: Method function.
        :return: Custom reaction object.
        """
        if isinstance(func_, CustomReaction):
            if func_.priority == priority:
                return func_
            func_ = func_.func
        return CustomReaction(func_, priority=priority)

    if func is not None:
        return _reaction(func)
    else:
        return _reaction


@final
class CustomReaction(BaseReaction):
    """
    Custom method-like that gets called whenever an action is sent through the object.

    Inherits from:
      - :class:`objetto.bases.BaseReaction`

    :param func: Function.
    :type func: function

    :param priority: Priority.
    :type priority: int or None
    """

    __slots__ = ("__func", "__priority")

    def __init__(self, func, priority=None):
        # type: (Callable[[BaseObject, Action, Phase], None], Optional[int]) -> None
        super(CustomReaction, self).__init__()

        # 'func'
        with ReraiseContext(TypeError):
            assert_is_callable(func)

        # 'priority'
        if priority is not None:
            with ReraiseContext(TypeError, "'priority' parameter"):
                assert_is_instance(priority, integer_types)

        self.__func = func
        self._priority = priority

    def __call__(self, obj, action, phase):
        """
        Run function.

        :param obj: Object.
        :type obj: objetto.bases.BaseObject

        :param action: Action.
        :type action: objetto.objects.Action

        :param phase: Phase.
        :type phase: `objetto.constants.PRE` or :data:`objetto.constants.POST`
        """
        self.func(obj, action, phase)

    def to_dict(self):
        # type: () -> Dict[str, Any]
        """
        Convert to dictionary.

        :return: Dictionary.
        :rtype: dict[str, Any]
        """
        dct = super(CustomReaction, self).to_dict()
        dct.update(
            {
                "func": self.func,
            }
        )
        return dct

    @property
    def func(self):
        # type: () -> Callable[[BaseObject, Action, Phase], None]
        """
        Function.

        :rtype: function
        """
        return self.__func


class UniqueAttributes(BaseReaction):
    """
    Asserts that children have unique attributes within a collection.
    Initialize with attribute names and optional incrementer functions.

    Inherits from:
      - :class:`objetto.bases.BaseReaction`

    :param names: Attribute names.
    :type names: str

    :param incrementers: Incrementer functions.
    :type incrementers: function
    """

    __slots__ = ("__names", "__incrementers")

    def __init__(self, *names, **incrementers):
        # type: (str, Callable[[Any, FrozenSet[Any]], Any]) -> None
        super(UniqueAttributes, self).__init__()

        # Check names.
        for name in names:
            assert_is_instance(name, string_types)
        # noinspection PyTypeChecker
        all_names = frozenset(names).union(incrementers)
        if not all_names:
            error = "no attribute names provided"
            raise ValueError(error)

        # Check incrementers.
        # noinspection PyTypeChecker
        all_incrementers = DictState(
            incrementers
        )  # type: DictState[str, Callable[[Any, FrozenSet[Any]], Any]]
        for name, incrementer in iteritems(all_incrementers):
            if not callable(incrementer):
                error = "expected a callable incrementer for '{}', got '{}'".format(
                    name, type(incrementer).__name__
                )
                raise TypeError(error)

        self.__names = all_names  # type: FrozenSet[str]
        self.__incrementers = all_incrementers

    def __call__(self, obj, action, phase):
        # type: (BaseObject, Action, Phase) -> None
        """
        React to new children or children's attribute changes.

        :param obj: Object.
        :type obj: objetto.bases.BaseObject

        :param action: Action.
        :type action: objetto.objects.Action

        :param phase: Phase.
        :type phase: `objetto.constants.PRE` or :data:`objetto.constants.POST`
        """

        # Ignore non-atomic changes.
        if not isinstance(action.change, BaseAtomicChange):
            return

        # Adopting or releasing children.
        if not action.locations:

            # Old children.
            if action.change.old_children:

                # After releasing, update metadata.
                if phase is phase.POST:
                    with obj.app.__.update_metadata_context(obj) as (read, update):

                        # Get existing cache.
                        metadata = read()
                        if UNIQUE_ATTRIBUTES_METADATA_KEY not in metadata:
                            cache = InteractiveDictData(
                                ()
                            )  # type: InteractiveDictData[str, Any]
                        else:
                            cache = metadata[UNIQUE_ATTRIBUTES_METADATA_KEY]

                        # For every old child, remove values from cache.
                        for child in action.change.old_children:
                            if not isinstance(child, Object):
                                continue
                            for name in self.__names:
                                if hasattr(child, name):
                                    value = getattr(child, name)
                                    try:
                                        hash(value)
                                    except TypeError:
                                        continue
                                    if name in cache:
                                        cache = cache.set(
                                            name, cache[name].discard(value)
                                        )

                        # Update metadata.
                        update({UNIQUE_ATTRIBUTES_METADATA_KEY: cache})

            # New children.
            if action.change.new_children:

                # Before adopting.
                if phase is Phase.PRE:
                    children = frozenset(
                        c for c in action.change.new_children if isinstance(c, Object)
                    )
                    all_new_values = self.__react(obj, children)

                    # Update children with new values.
                    for child, new_values in iteritems(all_new_values):
                        child.update(new_values)

                # After adopting, update metadata.
                elif phase is phase.POST:
                    with obj.app.__.update_metadata_context(obj) as (read, update):

                        # Get existing cache.
                        metadata = read()
                        if UNIQUE_ATTRIBUTES_METADATA_KEY not in metadata:
                            cache = InteractiveDictData()
                        else:
                            cache = metadata[UNIQUE_ATTRIBUTES_METADATA_KEY]

                        # For every new child, add values to cache.
                        for child in action.change.new_children:
                            if not isinstance(child, Object):
                                continue
                            for name in self.__names:
                                if hasattr(child, name):
                                    value = getattr(child, name)
                                    try:
                                        hash(value)
                                    except TypeError:
                                        continue
                                    if name not in cache:
                                        cache = cache.set(
                                            name,
                                            InteractiveDictData({value: child}),
                                        )
                                    else:
                                        cache = cache.set(
                                            name, cache[name].set(value, child)
                                        )

                        # Update metadata.
                        update({UNIQUE_ATTRIBUTES_METADATA_KEY: cache})

        # Changes in existing children.
        elif len(action.locations) == 1:
            if isinstance(action.change, Update):

                # Before changes.
                if phase is Phase.PRE:
                    child = cast("Object", action.sender)
                    children = frozenset((child,))
                    all_new_values = self.__react(
                        obj, children, {child: action.change.new_values}
                    )
                    if child in all_new_values:
                        new_values = all_new_values[child]
                        new_input_values = dict(
                            (name, new_values.get(name, value))
                            for name, value in iteritems(action.change.new_values)
                        )

                        # Callback to update children with new values.
                        def callback():
                            child.update(new_input_values)

                        # Reject changes and run callback.
                        raise RejectChangeException(
                            "attributes not unique", action.change, callback
                        )

                # After changes, update metadata.
                elif phase is Phase.POST:
                    with obj.app.__.update_metadata_context(obj) as (read, update):

                        # Get existing cache.
                        metadata = read()
                        if UNIQUE_ATTRIBUTES_METADATA_KEY not in metadata:
                            cache = InteractiveDictData()
                        else:
                            cache = metadata[UNIQUE_ATTRIBUTES_METADATA_KEY]

                        # Update cache.
                        child = action.sender
                        if isinstance(child, Object):
                            for name, new_value in iteritems(action.change.new_values):
                                old_value = action.change.old_values.get(name, MISSING)
                                if name not in cache:
                                    cache = cache.set(
                                        name,
                                        InteractiveDictData({new_value: child}),
                                    )
                                else:
                                    cache = cache.set(
                                        name,
                                        cache[name]
                                        .discard(old_value)
                                        .set(new_value, child),
                                    )

                            # Update metadata.
                            update({UNIQUE_ATTRIBUTES_METADATA_KEY: cache})

    def __react(
        self,
        obj,  # type: BaseObject
        children,  # type: FrozenSet[Object]
        child_new_values=None,  # type: Optional[Dict[Object, Mapping[str, Any]]]
    ):
        # type: (...) -> Dict[Object, Dict[str, Any]]
        """React and return new values."""

        # Get children and existing children.
        children = frozenset(c for c in children if isinstance(c, Object))
        existing_children = frozenset(c for c in obj._children if isinstance(c, Object))
        _child_new_values = child_new_values or {}

        # Build a dictionary with existing values.
        all_values = defaultdict(ValueCounter)  # type: Dict[str, Counter[Any]]
        for child in existing_children.union(children):

            # Get value.
            for name in self.names:

                # Get new value if existing child is changing.
                if child in _child_new_values:
                    if name in _child_new_values[child]:
                        all_values[name][_child_new_values[child][name]] += 1
                        continue

                # Get existing value.
                all_values[name][getattr(child, name)] += 1

        # Prepare dict to store new values.
        all_new_values = defaultdict(dict)  # type: Dict[Object, Dict[str, Any]]

        # Check children's values for collisions.
        for child in children:
            for name in self.names:

                # If specifying children new values, skip the ones not specified.
                if child_new_values is not None:
                    if child not in child_new_values:
                        continue
                    if name not in child_new_values[child]:
                        continue
                    value = child_new_values[child][name]
                else:
                    # Get existing values and child's value for this attribute.
                    value = getattr(child, name)

                # Get all values for this attribute.
                values, values_counter = (
                    frozenset(v for v in all_values[name] if all_values[name][v]),
                    all_values[name],
                )

                # There's a collision.
                if values_counter.get(value, 0) > 1:

                    # Prepare error message.
                    error = ("another object already has '{}' set to {}").format(
                        name, repr(value)
                    )

                    # Try to use incrementer to get a unique value.
                    if name in self.incrementers:
                        incrementer = self.incrementers[name]
                        new_value = incrementer(value, values)

                        # Incrementer was able to provide a unique value, so
                        # keep track of it and update all values dictionary.
                        all_values[name][value] -= 1
                        if values_counter.get(new_value, 0) < 2:
                            all_new_values[child][name] = new_value
                            all_values[name][new_value] += 1
                            continue

                        # Incrementer failed to make it unique, add to the error
                        # message and error out as a runtime error.
                        else:
                            error += (
                                ", incrementer {} failed to make it unique "
                                "(result was {})"
                            ).format(incrementer, repr(new_value))
                            raise RuntimeError(error)

                    # Not unique, raise value error.
                    raise ValueError(error)

        return all_new_values

    def to_dict(self):
        # type: () -> Dict[str, Any]
        """
        Convert to dictionary.

        :return: Dictionary.
        :rtype: dict[str, Any]
        """
        dct = super(UniqueAttributes, self).to_dict()
        dct.update(
            {
                "names": self.names,
                "incrementers": self.incrementers,
            }
        )
        return dct

    @property
    def names(self):
        # type: () -> FrozenSet[str]
        """
        Names.

        :rtype: tuple[str]
        """
        return self.__names

    @property
    def incrementers(self):
        # type: () -> DictState[str, Callable]
        """
        Incrementer functions.

        :rtype: objetto.states.DictState[str, function]
        """
        return self.__incrementers


class LimitChildren(BaseReaction):
    """
    Limit the number of children.

    Inherits from:
      - :class:`objetto.bases.BaseReaction`

    :param minimum: Minimum.
    :type minimum: int or None

    :param maximum: Maximum.
    :type maximum: int or None
    """

    __slots__ = ("__minimum", "__maximum")

    def __init__(self, minimum=None, maximum=None):
        # type: (Optional[int], Optional[int]) -> None
        super(LimitChildren, self).__init__()

        if minimum is not None:
            minimum = int(minimum)
            if minimum < 0:
                error = "minimum cannot be less than zero"
                raise ValueError(error)
        if maximum is not None:
            maximum = int(maximum)

        self.__minimum = minimum
        self.__maximum = maximum

    def __call__(self, obj, action, phase):
        # type: (BaseObject, Action, Phase) -> None
        """
        React to atomic changes.

        :param obj: Object.
        :type obj: objetto.bases.BaseObject

        :param action: Action.
        :type action: objetto.objects.Action

        :param phase: Phase.
        :type phase: `objetto.constants.PRE` or :data:`objetto.constants.POST`
        """

        # Ignore non-atomic changes.
        if not isinstance(action.change, BaseAtomicChange):
            return

        if len(action.locations) == 0:
            if phase is Phase.PRE:
                if action.change.new_children or action.change.old_children:
                    current_len = len(obj._children)
                    new_len = current_len + (
                        len(action.change.new_children)
                        - len(action.change.old_children)
                    )

                    # Growing, check for maximum.
                    if self.maximum is not None and new_len > self.maximum:
                        error_msg = (
                            "tried to add too many children (maximum is {})"
                        ).format(self.maximum)
                        raise ValueError(error_msg)

                    # Shrinking, check for minimum.
                    elif self.minimum is not None and current_len >= self.minimum:
                        if new_len < self.minimum:
                            history = obj._history
                            if history is None or not history.undoing:
                                error_msg = (
                                    "tried to remove too many children (minimum is {})"
                                ).format(self.minimum)
                                raise ValueError(error_msg)

    def to_dict(self):
        # type: () -> Dict[str, Any]
        """
        Convert to dictionary.

        :return: Dictionary.
        :rtype: dict[str, Any]
        """
        dct = super(LimitChildren, self).to_dict()
        dct.update(
            {
                "minimum": self.minimum,
                "maximum": self.maximum,
            }
        )
        return dct

    @property
    def minimum(self):
        # type: () -> Optional[int]
        """
        Minimum.

        :rtype: int or None
        """
        return self.__minimum

    @property
    def maximum(self):
        # type: () -> Optional[int]
        """
        Maximum.

        :rtype: int or None
        """
        return self.__maximum


class Limit(BaseReaction):
    """
    Limit the number of values.

    Inherits from:
      - :class:`objetto.bases.BaseReaction`

    :param minimum: Minimum.
    :type minimum: int or None

    :param maximum: Maximum.
    :type maximum: int or None
    """

    __slots__ = ("__minimum", "__maximum")

    def __init__(self, minimum=None, maximum=None):
        # type: (Optional[int], Optional[int]) -> None
        super(Limit, self).__init__()

        if minimum is not None:
            minimum = int(minimum)
            if minimum < 0:
                error = "minimum cannot be less than zero"
                raise ValueError(error)
        if maximum is not None:
            maximum = int(maximum)

        self.__minimum = minimum
        self.__maximum = maximum

    def __call__(self, obj, action, phase):
        # type: (BaseObject, Action, Phase) -> None
        """
        React to atomic changes.

        :param obj: Object.
        :param action: Action.
        :param phase: Phase.
        """

        # Ignore non-atomic changes.
        if not isinstance(action.change, BaseAtomicChange):
            return

        if len(action.locations) == 0:
            if phase is Phase.PRE:
                current_len = len(action.change.old_state)
                new_len = len(action.change.new_state)

                # Growing, check for maximum.
                if self.maximum is not None and new_len > self.maximum:
                    error_msg = ("tried to add too many values (maximum is {})").format(
                        self.maximum
                    )
                    raise ValueError(error_msg)

                # Shrinking, check for minimum.
                elif self.minimum is not None and current_len >= self.minimum:
                    if new_len < self.minimum:
                        history = obj._history
                        if history is None or not history.undoing:
                            error_msg = (
                                "tried to remove too many values (minimum is {})"
                            ).format(self.minimum)
                            raise ValueError(error_msg)

    def to_dict(self):
        # type: () -> Dict[str, Any]
        """
        Convert to dictionary.

        :return: Dictionary.
        :rtype: dict[str, Any]
        """
        dct = super(Limit, self).to_dict()
        dct.update(
            {
                "minimum": self.minimum,
                "maximum": self.maximum,
            }
        )
        return dct

    @property
    def minimum(self):
        # type: () -> Optional[int]
        """
        Minimum.

        :rtype: int or None
        """
        return self.__minimum

    @property
    def maximum(self):
        # type: () -> Optional[int]
        """
        Maximum.

        :rtype: int or None
        """
        return self.__maximum



# -*- coding: utf-8 -*-
"""History object."""

from ._history import BatchChanges, HistoryObject
from ._objects.bases import HistoryDescriptor
from .objects import history_descriptor

__all__ = ["history_descriptor", "HistoryDescriptor", "HistoryObject", "BatchChanges"]



# -*- coding: utf-8 -*-
import pytest

from objetto.utils.subject_observer import Observer, ObserverToken, Subject


def test_send_payload():
    class MyObserver(Observer):
        def __init__(self):
            self.payload = ()

        def __observe__(self, *_payload):
            self.payload = payload

    subject = Subject()

    observer_a = MyObserver()
    observer_a.start_observing(subject)

    observer_b = MyObserver()
    observer_b.start_observing(subject)

    payload = 1, 2, 3
    errors = subject.send(*payload)

    assert not errors
    assert observer_a.payload == payload
    assert observer_b.payload == payload

    observer_a.stop_observing(subject)
    observer_b.stop_observing(subject)

    other_payload = 4, 5, 6
    errors = subject.send(*other_payload)

    assert not errors
    assert observer_a.payload == payload
    assert observer_b.payload == payload


def test_tokens():
    class MyObserver(Observer):
        def __init__(self, index):
            self.payload = ()
            self.index = index

        def __observe__(self, *_payload):
            called.append(self.index)
            if self.index < 99:
                wait_for = self.index + 1
                tokens[wait_for].wait()
            observed.append(self.index)

    subject = Subject()

    observers = []
    tokens = []
    for i in range(100):
        observer = MyObserver(i)
        tokens.append(observer.start_observing(subject))
        observers.append(observer)

    called = []
    observed = []
    payload = 1, 2, 3
    errors = subject.send(payload)

    assert sorted(called) == list(range(100))
    assert len(called) == 100
    assert not errors
    assert observed == list(reversed(range(100)))


def test_token_cycle_errors():
    class MyObserver(Observer):
        def __init__(self, index):
            self.payload = ()
            self.index = index

        def __observe__(self, *_payload):
            called.append(self.index)
            tokens[99 - self.index].wait()
            observed.append(self.index)

    subject = Subject()

    observers = []
    tokens = []
    for i in range(100):
        observer = MyObserver(i)
        tokens.append(observer.start_observing(subject))
        observers.append(observer)

    called = []
    observed = []
    payload = 1, 2, 3
    errors = subject.send(payload)

    assert sorted(called) == list(range(100))
    assert len(called) == 100
    assert len(errors) == 100
    assert not observed


def test_prevent_token_instantiation():
    with pytest.raises(RuntimeError):
        subject, observer = Subject(), Observer()
        ObserverToken(subject, observer)


if __name__ == "__main__":
    pytest.main()



# -*- coding: utf-8 -*-
"""Immutable dict state."""

from typing import TYPE_CHECKING, TypeVar, cast, overload

try:
    import collections.abc as collections_abc
except ImportError:
    import collections as collections_abc  # type: ignore

from pyrsistent import pmap
from six import iteritems, iterkeys, itervalues

from .._bases import BaseInteractiveDict, final
from ..utils.custom_repr import custom_mapping_repr
from ..utils.recursive_repr import recursive_repr
from .bases import BaseState

if TYPE_CHECKING:
    from typing import Any, Iterable, Iterator, Mapping, Tuple, Type, Union

    from pyrsistent.typing import PMap

__all__ = ["DictState"]


KT = TypeVar("KT")  # Key type.
VT = TypeVar("VT")  # Value type.


# noinspection PyTypeChecker
_DS = TypeVar("_DS", bound="DictState")


@final
class DictState(BaseState[KT], BaseInteractiveDict[KT, VT]):
    """
    Immutable dictionary state.

    Inherits from:
      - :class:`objetto.bases.BaseState`
      - :class:`objetto.bases.BaseInteractiveDict`

    :param initial: Initial values.
    :type initial: collections.abc.Mapping or collections.abc.Iterable[\
tuple[collections.abc.Hashable, Any]]
    """

    __slots__ = ()

    @classmethod
    def _make(cls, internal=pmap()):
        # type: (Type[_DS], PMap[KT, VT]) -> _DS
        """
        Make new state by directly setting the internal state.

        :param internal: Internal state.
        :return: State.
        """
        return super(DictState, cls)._make(internal)

    @staticmethod
    def _make_internal(initial):
        # type: (Union[Mapping[KT, VT], Iterable[Tuple[KT, VT]]]) -> PMap[KT, VT]
        """
        Initialize internal state.

        :param initial: Initial values.
        """
        return pmap(initial)

    def __init__(self, initial=()):
        # type: (Union[Mapping[KT, VT], Iterable[Tuple[KT, VT]]]) -> None
        super(DictState, self).__init__(initial=initial)

    def __hash__(self):
        # type: () -> int
        """
        Get hash.

        :return: Hash.
        :rtype: int
        """
        return super(DictState, self).__hash__()

    def __eq__(self, other):
        # type: (object) -> bool
        """
        Compare for equality.

        :param other: Another object.

        :return: True if equal.
        :rtype: bool
        """
        if self is other:
            return True
        if not isinstance(other, collections_abc.Hashable):
            return self._internal == other
        if isinstance(other, DictState):
            return self._internal == other._internal
        return False

    def __contains__(self, key):
        # type: (Any) -> bool
        """
        Get whether key is present.

        :param key: Key.
        :type key: collections.abc.Hashable

        :return: True if contains.
        :rtype: bool
        """
        return key in self._internal

    def __iter__(self):
        # type: () -> Iterator[KT]
        """
        Iterate over keys.

        :return: Keys iterator.
        :rtype: collections.abc.Iterator[collections.abc.Hashable]
        """
        for key in iterkeys(self._internal):
            yield key

    def __len__(self):
        # type: () -> int
        """
        Get key count.

        :return: Key count.
        :rtype: int
        """
        return len(self._internal)

    @recursive_repr
    def __repr__(self):
        # type: () -> str
        """
        Get representation.

        :return: Representation.
        :rtype: str
        """
        return custom_mapping_repr(
            self._internal,
            prefix="{}({{".format(type(self).__fullname__),
            suffix="})",
            sorting=True,
            sort_key=lambda i: hash(i[0]),
        )

    def __reversed__(self):
        # type: () -> Iterator[KT]
        """
        Iterate over reversed keys.

        :return: Reversed keys iterator.
        :rtype: collections.abc.Iterator[collections.abc.Hashable]
        """
        return reversed(list(self.__iter__()))

    def __getitem__(self, key):
        # type: (KT) -> VT
        """
        Get value for key.

        :param key: Key.
        :type key: collections.abc.Hashable

        :return: Value.

        :raises KeyError: Key is not present.
        """
        return self._internal[key]

    def _clear(self):
        # type: (_DS) -> _DS
        """
        Clear.

        :return: Transformed.
        :rtype: objetto.states.DictState
        """
        return self._make()

    def _discard(self, key):
        # type: (_DS, KT) -> _DS
        """
        Discard key if it exists.

        :param key: Key.
        :type key: collections.abc.Hashable

        :return: Transformed.
        :rtype: objetto.states.DictState
        """
        return self._make(self._internal.discard(key))

    def _remove(self, key):
        # type: (_DS, KT) -> _DS
        """
        Delete existing key.

        :param key: Key.
        :type key: collections.abc.Hashable

        :return: Transformed.
        :rtype: objetto.states.DictState

        :raises KeyError: Key is not present.
        """
        return self._make(self._internal.remove(key))

    def _set(self, key, value):
        # type: (_DS, KT, VT) -> _DS
        """
        Set value for key.

        :param key: Key.
        :type key: collections.abc.Hashable

        :param value: Value.

        :return: Transformed.
        :rtype: objetto.states.DictState
        """
        return self._make(self._internal.set(key, value))

    @overload
    def _update(self, __m, **kwargs):
        # type: (_DS, Mapping[KT, VT], VT) -> _DS
        pass

    @overload
    def _update(self, __m, **kwargs):
        # type: (_DS, Iterable[Tuple[KT, VT]], VT) -> _DS
        pass

    @overload
    def _update(self, **kwargs):
        # type: (_DS, VT) -> _DS
        pass

    def _update(self, *args, **kwargs):
        """
        Update keys and values.
        Same parameters as :meth:`dict.update`.

        :return: Transformed.
        :rtype: objetto.states.DictState
        """
        return self._make(self._internal.update(dict(*args, **kwargs)))

    def get(self, key, fallback=None):
        # type: (KT, Any) -> Union[VT, Any]
        """
        Get value for key, return fallback value if key is not present.

        :param key: Key.
        :type key: collections.abc.Hashable

        :param fallback: Fallback value.

        :return: Value or fallback value.
        """
        return self._internal.get(key, fallback)

    def iteritems(self):
        # type: () -> Iterator[Tuple[KT, VT]]
        """
        Iterate over items.

        :return: Items iterator.
        :rtype: collections.abc.Iterator[tuple[collections.abc.Hashable, Any]]
        """
        for key, value in iteritems(self._internal):
            yield key, value

    def iterkeys(self):
        # type: () -> Iterator[KT]
        """
        Iterate over keys.

        :return: Keys iterator.
        :rtype: collections.abc.Iterator[collections.abc.Hashable]
        """
        for key in iterkeys(self._internal):
            yield key

    def itervalues(self):
        # type: () -> Iterator[VT]
        """
        Iterate over values.

        :return: Values iterator.
        :rtype: collections.abc.Iterator
        """
        for value in itervalues(self._internal):
            yield value

    def find_with_attributes(self, **attributes):
        # type: (Any) -> VT
        """
        Find first value that matches unique attribute values.

        :param attributes: Attributes to match.

        :return: Value that has matching attributes.

        :raises ValueError: No attributes provided or no match found.
        """
        if not attributes:
            error = "no attributes provided"
            raise ValueError(error)
        for value in itervalues(self._internal):
            for a_name, a_value in iteritems(attributes):
                if not hasattr(value, a_name) or getattr(value, a_name) != a_value:
                    break
            else:
                return value
        error = "could not find a match for {}".format(
            custom_mapping_repr(
                attributes,
                prefix="(",
                template="{key}={value}",
                suffix=")",
                key_repr=str,
            ),
        )
        raise ValueError(error)

    @property
    def _internal(self):
        # type: () -> PMap[KT, VT]
        """Internal values."""
        return cast("PMap[KT, VT]", super(DictState, self)._internal)



# -*- coding: utf-8 -*-
"""Base set classes."""

from abc import abstractmethod
from typing import TYPE_CHECKING, TypeVar

try:
    import collections.abc as collections_abc
except ImportError:
    import collections as collections_abc  # type: ignore

from slotted import SlottedMutableSet, SlottedSet

from .bases import (
    BaseCollection,
    BaseInteractiveCollection,
    BaseMutableCollection,
    BaseProtectedCollection,
    Generic,
    final,
)

if TYPE_CHECKING:
    from typing import AbstractSet, Iterable

__all__ = [
    "BaseSet",
    "BaseProtectedSet",
    "BaseInteractiveSet",
    "BaseMutableSet",
]


T = TypeVar("T")  # Any type.
T_co = TypeVar("T_co", covariant=True)  # Any type covariant containers.


class BaseSet(SlottedSet, BaseCollection[T_co], Generic[T_co]):
    """
    Base set collection.

    Inherits from:
      - :class:`objetto.bases.BaseCollection`
      - :class:`slotted.SlottedSet`
      - :class:`typing.Generic`

    Inherited By:
      - :class:`objetto.bases.BaseProtectedSet`
    """

    __slots__ = ()

    def __hash__(self):
        """
        Prevent hashing (not hashable by default).

        :raises TypeError: Not hashable.
        """
        error = "unhashable type: '{}'".format(type(self).__fullname__)
        raise TypeError(error)

    @final
    def __le__(self, other):
        # type: (AbstractSet) -> bool
        """
        Less equal operator (self <= other).

        :param other: Another set or any object.
        :type other: collections.abc.Set

        :return: True if considered less equal.
        :rtype: bool
        """
        if not isinstance(other, collections_abc.Set):
            return NotImplemented
        if type(other) not in (set, frozenset):
            other = set(other)
        return set(self).__le__(other)

    @final
    def __lt__(self, other):
        # type: (AbstractSet) -> bool
        """
        Less than operator: `self < other`.

        :param other: Another set or any object.
        :type other: collections.abc.Set

        :return: True if considered less than.
        :rtype: bool
        """
        if not isinstance(other, collections_abc.Set):
            return NotImplemented
        if type(other) not in (set, frozenset):
            other = set(other)
        return set(self).__lt__(other)

    @final
    def __gt__(self, other):
        # type: (AbstractSet) -> bool
        """
        Greater than operator: `self > other`.

        :param other: Another set or any object.
        :type other: collections.abc.Set

        :return: True if considered greater than.
        :rtype: bool
        """
        if not isinstance(other, collections_abc.Set):
            return NotImplemented
        if type(other) not in (set, frozenset):
            other = set(other)
        return set(self).__gt__(other)

    @final
    def __ge__(self, other):
        # type: (AbstractSet) -> bool
        """
        Greater equal operator: `self >= other`.

        :param other: Another set or any object.
        :type other: collections.abc.Set

        :return: True if considered greater equal.
        :rtype: bool
        """
        if not isinstance(other, collections_abc.Set):
            return NotImplemented
        if type(other) not in (set, frozenset):
            other = set(other)
        return set(self).__ge__(other)

    @final
    def __and__(self, other):
        # type: (Iterable) -> BaseSet
        """
        Get intersection: `self & other`.

        :param other: Iterable or any other object.
        :type other: collections.abc.Iterable

        :return: Intersection or `NotImplemented` if not an iterable.
        :rtype: objetto.bases.BaseSet
        """
        if not isinstance(other, collections_abc.Iterable):
            return NotImplemented
        return self.intersection(other)

    @final
    def __rand__(self, other):
        # type: (Iterable) -> BaseSet
        """
        Get intersection: `other & self`.

        :param other: Iterable or any other object.
        :type other: collections.abc.Iterable

        :return: Intersection or `NotImplemented` if not an iterable.
        :rtype: objetto.bases.BaseSet
        """
        return self.__and__(other)

    @final
    def __sub__(self, other):
        # type: (Iterable) -> BaseSet
        """
        Get difference: `self - other`.

        :param other: Iterable or any other object.
        :type other: collections.abc.Iterable

        :return: Difference or `NotImplemented` if not an iterable.
        :rtype: objetto.bases.BaseSet
        """
        if not isinstance(other, collections_abc.Iterable):
            return NotImplemented
        return self.difference(other)

    @final
    def __rsub__(self, other):
        # type: (Iterable) -> BaseSet
        """
        Get inverse difference: `other - self`.

        :param other: Iterable or any other object.
        :type other: collections.abc.Iterable

        :return: Inverse difference or `NotImplemented` if not an iterable.
        :rtype: objetto.bases.BaseSet
        """
        if not isinstance(other, collections_abc.Iterable):
            return NotImplemented
        return self.inverse_difference(other)

    @final
    def __or__(self, other):
        # type: (Iterable) -> BaseSet
        """
        Get union: `self | other`.

        :param other: Iterable or any other object.
        :type other: collections.abc.Iterable

        :return: Union or `NotImplemented` if not an iterable.
        :rtype: objetto.bases.BaseSet
        """
        if not isinstance(other, collections_abc.Iterable):
            return NotImplemented
        return self.union(other)

    @final
    def __ror__(self, other):
        # type: (Iterable) -> BaseSet
        """
        Get union: `other | self`.

        :param other: Iterable or any other object.
        :type other: collections.abc.Iterable

        :return: Union or `NotImplemented` if not an iterable.
        :rtype: objetto.bases.BaseSet
        """
        return self.__or__(other)

    @final
    def __xor__(self, other):
        # type: (Iterable) -> BaseSet
        """
        Get symmetric difference: `self ^ other`.

        :param other: Iterable or any other object.
        :type other: collections.abc.Iterable

        :return: Symmetric difference or `NotImplemented` if not an iterable.
        :rtype: objetto.bases.BaseSet
        """
        if not isinstance(other, collections_abc.Iterable):
            return NotImplemented
        return self.symmetric_difference(other)

    @final
    def __rxor__(self, other):
        # type: (Iterable) -> BaseSet
        """
        Get symmetric difference: `other ^ self`.

        :param other: Iterable or any other object.
        :type other: collections.abc.Iterable

        :return: Symmetric difference or `NotImplemented` if not an iterable.
        :rtype: objetto.bases.BaseSet
        """
        return self.__xor__(other)

    @abstractmethod
    def __eq__(self, other):
        # type: (object) -> bool
        """
        Compare for equality.

        :param other: Another object.

        :return: True if equal.
        :rtype: bool

        :raises NotImplementedError: Abstract method not implemented.
        """
        raise NotImplementedError()

    @classmethod
    @abstractmethod
    def _from_iterable(cls, iterable):
        # type: (Iterable) -> BaseSet
        """
        Make set from iterable.

        :param iterable: Iterable.
        :type iterable: collections.abc.Iterable

        :return: Set.
        :raises NotImplementedError: Abstract method not implemented.
        """
        raise NotImplementedError()

    @abstractmethod
    def _hash(self):
        # type: () -> int
        """
        Get hash.

        :return: Hash.
        :rtype: int

        :raises NotImplementedError: Abstract method not implemented.
        """
        raise NotImplementedError()

    @abstractmethod
    def isdisjoint(self, iterable):
        # type: (Iterable) -> bool
        """
        Get whether is a disjoint set of an iterable.

        :param iterable: Iterable.
        :type iterable: collections.abc.Iterable

        :return: True if is disjoint.
        :rtype: bool

        :raises NotImplementedError: Abstract method not implemented.
        """
        raise NotImplementedError()

    @abstractmethod
    def issubset(self, iterable):
        # type: (Iterable) -> bool
        """
        Get whether is a subset of an iterable.

        :param iterable: Iterable.
        :type iterable: collections.abc.Iterable

        :return: True if is subset.
        :rtype: bool

        :raises NotImplementedError: Abstract method not implemented.
        """
        raise NotImplementedError()

    @abstractmethod
    def issuperset(self, iterable):
        # type: (Iterable) -> bool
        """
        Get whether is a superset of an iterable.

        :param iterable: Iterable.
        :type iterable: collections.abc.Iterable

        :return: True if is superset.
        :rtype: bool

        :raises NotImplementedError: Abstract method not implemented.
        """
        raise NotImplementedError()

    @abstractmethod
    def intersection(self, iterable):
        # type: (Iterable) -> BaseSet
        """
        Get intersection.

        :param iterable: Iterable.
        :type iterable: collections.abc.Iterable

        :return: Intersection.
        :rtype: objetto.bases.BaseSet

        :raises NotImplementedError: Abstract method not implemented.
        """
        raise NotImplementedError()

    @abstractmethod
    def symmetric_difference(self, iterable):
        # type: (Iterable) -> BaseSet
        """
        Get symmetric difference.

        :param iterable: Iterable.
        :type iterable: collections.abc.Iterable

        :return: Symmetric difference.
        :rtype: objetto.bases.BaseSet

        :raises NotImplementedError: Abstract method not implemented.
        """
        raise NotImplementedError()

    @abstractmethod
    def union(self, iterable):
        # type: (Iterable) -> BaseSet
        """
        Get union.

        :param iterable: Iterable.
        :type iterable: collections.abc.Iterable

        :return: Union.
        :rtype: objetto.bases.BaseSet

        :raises NotImplementedError: Abstract method not implemented.
        """
        raise NotImplementedError()

    @abstractmethod
    def difference(self, iterable):
        # type: (Iterable) -> BaseSet
        """
        Get difference.

        :param iterable: Iterable.
        :type iterable: collections.abc.Iterable

        :return: Difference.
        :rtype: objetto.bases.BaseSet

        :raises NotImplementedError: Abstract method not implemented.
        """
        raise NotImplementedError()

    @abstractmethod
    def inverse_difference(self, iterable):
        # type: (Iterable) -> BaseSet
        """
        Get an iterable's difference to this.

        :param iterable: Iterable.
        :type iterable: collections.abc.Iterable

        :return: Inverse Difference.
        :rtype: objetto.bases.BaseSet

        :raises NotImplementedError: Abstract method not implemented.
        """
        raise NotImplementedError()


# noinspection PyCallByClass
type.__setattr__(BaseSet, "__hash__", None)  # force non-hashable


# noinspection PyTypeChecker
_BPS = TypeVar("_BPS", bound="BaseProtectedSet")


class BaseProtectedSet(BaseSet[T], BaseProtectedCollection[T]):
    """
    Base protected set collection.

    Inherits from:
      - :class:`objetto.bases.BaseSet`
      - :class:`objetto.bases.BaseProtectedCollection`

    Inherited By:
      - :class:`objetto.bases.BaseInteractiveSet`
      - :class:`objetto.bases.BaseMutableSet`
      - :class:`objetto.bases.BaseSetStructure`
    """

    __slots__ = ()

    @abstractmethod
    def _add(self, value):
        # type: (_BPS, T) -> _BPS
        """
        Add value.

        :param value: Value.

        :return: Transformed.
        :rtype: objetto.bases.BaseProtectedSet

        :raises NotImplementedError: Abstract method not implemented.
        """
        raise NotImplementedError()

    @abstractmethod
    def _discard(self, *values):
        # type: (_BPS, T) -> _BPS
        """
        Discard value(s).

        :param values: Value(s).
        :type values: collections.abc.Hashable

        :return: Transformed.
        :rtype: objetto.bases.BaseProtectedSet

        :raises ValueError: No values provided.
        :raises NotImplementedError: Abstract method not implemented.
        """
        raise NotImplementedError()

    @abstractmethod
    def _remove(self, *values):
        # type: (_BPS, T) -> _BPS
        """
        Remove existing value(s).

        :param values: Value(s).
        :type values: collections.abc.Hashable

        :return: Transformed.
        :rtype: objetto.bases.BaseProtectedSet

        :raises ValueError: No values provided.
        :raises KeyError: Value is not present.
        :raises NotImplementedError: Abstract method not implemented.
        """
        raise NotImplementedError()

    @abstractmethod
    def _replace(self, old_value, new_value):
        # type: (_BPS, T, T) -> _BPS
        """
        Replace existing value with a new one.

        :param old_value: Existing value.

        :param new_value: New value.

        :return: Transformed.
        :rtype: objetto.bases.BaseProtectedSet

        :raises KeyError: Value is not present.
        :raises NotImplementedError: Abstract method not implemented.
        """
        raise NotImplementedError()

    @abstractmethod
    def _update(self, iterable):
        # type: (_BPS, Iterable[T]) -> _BPS
        """
        Update with iterable.

        :param iterable: Iterable.
        :type iterable: collections.abc.Iterable[collections.abc.Hashable]

        :return: Transformed.
        :rtype: objetto.bases.BaseProtectedSet

        :raises NotImplementedError: Abstract method not implemented.
        """
        raise NotImplementedError()


# noinspection PyTypeChecker
_BIS = TypeVar("_BIS", bound="BaseInteractiveSet")


# noinspection PyAbstractClass
class BaseInteractiveSet(BaseProtectedSet[T], BaseInteractiveCollection[T]):
    """
    Base interactive set collection.

    Inherits from:
      - :class:`objetto.bases.BaseProtectedSet`
      - :class:`objetto.bases.BaseInteractiveCollection`

    Inherited By:
      - :class:`objetto.bases.BaseInteractiveSetStructure`
      - :class:`objetto.states.SetState`
    """

    __slots__ = ()

    @final
    def add(self, value):
        # type: (_BIS, T) -> _BIS
        """
        Add value.

        :param value: Value.

        :return: Transformed.
        :rtype: objetto.bases.BaseInteractiveSet
        """
        return self._add(value)

    @final
    def discard(self, *values):
        # type: (_BIS, T) -> _BIS
        """
        Discard value(s).

        :param values: Value(s).
        :type values: collections.abc.Hashable

        :return: Transformed.
        :rtype: objetto.bases.BaseInteractiveSet

        :raises ValueError: No values provided.
        """
        return self._discard(*values)

    @final
    def remove(self, *values):
        # type: (_BIS, T) -> _BIS
        """
        Remove existing value(s).

        :param values: Value(s).
        :type values: collections.abc.Hashable

        :return: Transformed.
        :rtype: objetto.bases.BaseInteractiveSet

        :raises ValueError: No values provided.
        :raises KeyError: Value is not present.
        """
        return self._remove(*values)

    @final
    def replace(self, old_value, new_value):
        # type: (_BIS, T, T) -> _BIS
        """
        Replace existing value with a new one.

        :param old_value: Existing value.

        :param new_value: New value.

        :return: Transformed.
        :rtype: objetto.bases.BaseInteractiveSet

        :raises KeyError: Old value is not present.
        """
        return self._replace(old_value, new_value)

    @final
    def update(self, iterable):
        # type: (_BIS, Iterable[T]) -> _BIS
        """
        Update with iterable.

        :param iterable: Iterable.
        :type iterable: collections.abc.Iterable

        :return: Transformed.
        :rtype: objetto.bases.BaseInteractiveSet
        """
        return self._update(iterable)


class BaseMutableSet(SlottedMutableSet, BaseProtectedSet[T], BaseMutableCollection[T]):
    """
    Base mutable set collection.

    Inherits from:
      - :class:`slotted.SlottedMutableSet`
      - :class:`objetto.bases.BaseProtectedSet`
      - :class:`objetto.bases.BaseMutableCollection`

    Inherited By:
      - :class:`objetto.bases.BaseMutableSetStructure`
      - :class:`objetto.objects.ProxySetObject`
    """

    __slots__ = ()

    @final
    def __iand__(self, iterable):
        """
        Intersect in place: `self &= iterable`.

        :param iterable: Iterable.
        :type iterable: collections.abc.Iterable

        :return: This mutable set.
        :rtype: objetto.bases.BaseMutableSet
        """
        self.intersection_update(iterable)
        return self

    @final
    def __isub__(self, iterable):
        """
        Difference in place: `self -= iterable`.

        :param iterable: Iterable.
        :type iterable: collections.abc.Iterable

        :return: This mutable set.
        :rtype: objetto.bases.BaseMutableSet
        """
        self.difference(iterable)
        return self

    @final
    def __ior__(self, iterable):
        """
        Update in place: `self |= iterable`.

        :param iterable: Iterable.
        :type iterable: collections.abc.Iterable

        :return: This mutable set.
        :rtype: objetto.bases.BaseMutableSet
        """
        self.update(iterable)
        return self

    @final
    def __ixor__(self, iterable):
        """
        Symmetric difference in place: `self ^= iterable`.

        :param iterable: Iterable.
        :type iterable: collections.abc.Iterable

        :return: This mutable set.
        :rtype: objetto.bases.BaseMutableSet
        """
        if iterable is self:
            self.clear()
        else:
            self.symmetric_difference_update(iterable)
        return self

    @abstractmethod
    def pop(self):
        # type: () -> T
        """
        Pop value.

        :return: Value.

        :raises KeyError: Empty set.
        :raises NotImplementedError: Abstract method not implemented.
        """
        raise NotImplementedError()

    @abstractmethod
    def intersection_update(self, iterable):
        # type: (Iterable[T]) -> None
        """
        Intersect.

        :param iterable: Iterable.
        :type iterable: collections.abc.Iterable

        :raises NotImplementedError: Abstract method not implemented.
        """
        raise NotImplementedError()

    @abstractmethod
    def symmetric_difference_update(self, iterable):
        # type: (Iterable[T]) -> None
        """
        Symmetric difference.

        :param iterable: Iterable.
        :type iterable: collections.abc.Iterable

        :raises NotImplementedError: Abstract method not implemented.
        """
        raise NotImplementedError()

    @abstractmethod
    def difference_update(self, iterable):
        # type: (Iterable[T]) -> None
        """
        Difference.

        :param iterable: Iterable.
        :type iterable: collections.abc.Iterable

        :raises NotImplementedError: Abstract method not implemented.
        """
        raise NotImplementedError()

    @final
    def clear(self):
        # type: () -> None
        """Clear."""
        self._clear()

    @final
    def add(self, value):
        # type: (T) -> None
        """
        Add value.

        :param value: Value.
        """
        self._add(value)

    @final
    def discard(self, *values):
        # type: (T) -> None
        """
        Discard value(s).

        :param values: Value(s).
        :type values: collections.abc.Hashable

        :raises ValueError: No values provided.
        """
        self._discard(*values)

    @final
    def remove(self, *values):
        # type: (T) -> None
        """
        Remove existing value(s).

        :param values: Value(s).
        :type values: collections.abc.Hashable

        :raises ValueError: No values provided.
        :raises KeyError: Value is not present.
        """
        self._remove(*values)

    @final
    def replace(self, old_value, new_value):
        # type: (T, T) -> None
        """
        Replace existing value with a new one.

        :param old_value: Existing value.

        :param new_value: New value.
        """
        self._replace(old_value, new_value)

    @final
    def update(self, iterable):
        # type: (Iterable[T]) -> None
        """
        Update with iterable.

        :param iterable: Iterable.
        :type iterable: collections.abc.Iterable
        """
        self._update(iterable)



# -*- coding: utf-8 -*-
"""Set structures."""

from abc import abstractmethod
from typing import TYPE_CHECKING, TypeVar

try:
    import collections.abc as collections_abc
except ImportError:
    import collections as collections_abc  # type: ignore

from six import with_metaclass

from .._bases import BaseInteractiveSet, BaseMutableSet, BaseProtectedSet, final
from .._states import SetState
from ..utils.custom_repr import custom_iterable_repr
from ..utils.recursive_repr import recursive_repr
from .bases import (
    BaseAuxiliaryStructure,
    BaseAuxiliaryStructureMeta,
    BaseInteractiveAuxiliaryStructure,
    BaseMutableAuxiliaryStructure,
)

if TYPE_CHECKING:
    from typing import Any, Iterable, Iterator


__all__ = [
    "BaseSetStructureMeta",
    "BaseSetStructure",
    "BaseInteractiveSetStructure",
    "BaseMutableSetStructure",
]


T = TypeVar("T")  # Any type.


# noinspection PyAbstractClass
class BaseSetStructureMeta(BaseAuxiliaryStructureMeta):
    """Metaclass for `SetStructure`."""


class BaseSetStructure(
    with_metaclass(
        BaseSetStructureMeta,
        BaseAuxiliaryStructure[T],
        BaseProtectedSet[T],
    )
):
    """
    Base set structure.

    Inherits from:
      - :class:`objetto.bases.BaseAuxiliaryStructure`
      - :class:`objetto.bases.BaseProtectedSet`

    Inherited By:
      - :class:`objetto.bases.BaseInteractiveSetStructure`
      - :class:`objetto.bases.BaseMutableSetStructure`
      - :class:`objetto.data.SetData`
      - :class:`objetto.objects.SetObject`
    """

    __slots__ = ()

    @recursive_repr
    def __repr__(self):
        # type: () -> str
        """
        Get representation.

        :return: Representation.
        :rtype: str
        """
        if type(self)._relationship.represented:
            return custom_iterable_repr(
                self._state,
                prefix="{}([".format(type(self).__fullname__),
                suffix="])",
                sorting=True,
                sort_key=lambda v: hash(v),
            )
        else:
            return "<{}>".format(type(self).__fullname__)

    @final
    def __len__(self):
        # type: () -> int
        """
        Get value count.

        :return: Value count.
        :rtype: int
        """
        return len(self._state)

    @final
    def __iter__(self):
        # type: () -> Iterator[T]
        """
        Iterate over values.

        :return: Values iterator.
        :rtype: collections.abc.Iterator[collections.abc.Hashable]
        """
        for value in self._state:
            yield value

    @final
    def __contains__(self, value):
        # type: (Any) -> bool
        """
        Get whether value is present.

        :param value: Value.
        :type value: collections.abc.Hashable

        :return: True if contains.
        :rtype: bool
        """
        return value in self._state

    def isdisjoint(self, iterable):
        # type: (Iterable) -> bool
        """
        Get whether is a disjoint set of an iterable.

        :param iterable: Iterable.
        :type iterable: collections.abc.Iterable

        :return: True if is disjoint.
        :rtype: bool
        """
        return self._state.isdisjoint(iterable)

    def issubset(self, iterable):
        # type: (Iterable) -> bool
        """
        Get whether is a subset of an iterable.

        :param iterable: Iterable.
        :type iterable: collections.abc.Iterable

        :return: True if is subset.
        :rtype: bool
        """
        return self._state.issubset(iterable)

    def issuperset(self, iterable):
        # type: (Iterable) -> bool
        """
        Get whether is a superset of an iterable.

        :param iterable: Iterable.
        :type iterable: collections.abc.Iterable

        :return: True if is superset.
        :rtype: bool
        """
        return self._state.issuperset(iterable)

    def intersection(self, iterable):
        # type: (Iterable) -> SetState
        """
        Get intersection.

        :param iterable: Iterable.
        :type iterable: collections.abc.Iterable

        :return: Intersection.
        :rtype: objetto.states.SetState
        """
        return self._state.intersection(iterable)

    def difference(self, iterable):
        # type: (Iterable) -> SetState
        """
        Get difference.

        :param iterable: Iterable.
        :type iterable: collections.abc.Iterable

        :return: Difference.
        :rtype: objetto.states.SetState
        """
        return self._state.difference(iterable)

    def inverse_difference(self, iterable):
        # type: (Iterable) -> SetState
        """
        Get an iterable's difference to this.

        :param iterable: Iterable.
        :type iterable: collections.abc.Iterable

        :return: Inverse Difference.
        :rtype: objetto.states.SetState
        """
        return self._state.inverse_difference(iterable)

    def symmetric_difference(self, iterable):
        # type: (Iterable) -> SetState
        """
        Get symmetric difference.

        :param iterable: Iterable.
        :type iterable: collections.abc.Iterable

        :return: Symmetric difference.
        :rtype: objetto.states.SetState
        """
        return self._state.symmetric_difference(iterable)

    def union(self, iterable):
        # type: (Iterable) -> SetState
        """
        Get union.

        :param iterable: Iterable.
        :type iterable: collections.abc.Iterable

        :return: Union.
        :rtype: objetto.states.SetState
        """
        return self._state.union(iterable)

    @property
    @abstractmethod
    def _state(self):
        # type: () -> SetState[T]
        """
        Internal state.

        :rtype: objetto.states.SetState

        :raises NotImplementedError: Abstract method not implemented.
        """
        raise NotImplementedError()


# noinspection PyAbstractClass
class BaseInteractiveSetStructure(
    BaseSetStructure[T],
    BaseInteractiveAuxiliaryStructure[T],
    BaseInteractiveSet[T],
):
    """
    Base interactive set structure.

    Inherits from:
      - :class:`objetto.bases.BaseSetStructure`
      - :class:`objetto.bases.BaseInteractiveAuxiliaryStructure`
      - :class:`objetto.bases.BaseInteractiveSet`

    Inherited By:
      - :class:`objetto.data.InteractiveSetData`
    """

    __slots__ = ()


# noinspection PyAbstractClass
class BaseMutableSetStructure(
    BaseMutableSet[T],
    BaseSetStructure[T],
    BaseMutableAuxiliaryStructure[T],
):
    """
    Base mutable set structure.

    Inherits from:
      - :class:`objetto.bases.BaseMutableSet`
      - :class:`objetto.bases.BaseSetStructure`
      - :class:`objetto.bases.BaseMutableAuxiliaryStructure`

    Inherited By:
      - :class:`objetto.objects.MutableSetObject`
    """

    __slots__ = ()



# -*- coding: utf-8 -*-

import pytest

from objetto.utils.list_operations import (
    pre_move,
    resolve_continuous_slice,
    resolve_index,
)


def test_resolve_index():
    my_list = ["a", "b", "c", "c"]
    length = len(my_list)

    for i in range(-4, 4):
        assert my_list[resolve_index(length, i)] == my_list[i]

    with pytest.raises(IndexError):
        resolve_index(length, 5)

    with pytest.raises(IndexError):
        resolve_index(length, -5)

    assert resolve_index(length, 5, clamp=True) == 4
    assert resolve_index(length, -5, clamp=True) == 0


def test_resolve_continuous_slice():
    my_list = ["a", "b", "c", "c"]
    length = len(my_list)

    assert resolve_continuous_slice(length, slice(0, 4)) == (0, 4)
    assert resolve_continuous_slice(length, slice(-5, 5)) == (0, 4)

    with pytest.raises(IndexError):
        resolve_continuous_slice(length, slice(0, 4, 2))


def test_pre_move():
    my_list = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, "a", "b", "c"]
    length = len(my_list)

    assert pre_move(length, 1, 2) is None
    assert pre_move(length, 2, 3) is None
    assert pre_move(length, 3, 0) == (3, 4, 0, 0)
    assert pre_move(length, 1, 3) == (1, 2, 3, 2)
    assert pre_move(length, slice(1, 3), 0) == (1, 3, 0, 0)
    assert pre_move(length, slice(1, 3), 4) == (1, 3, 4, 2)
    assert pre_move(length, slice(3, 5), 9) == (3, 5, 9, 7)


if __name__ == "__main__":
    pytest.main()



# -*- coding: utf-8 -*-
"""Weak reference type."""

from typing import TYPE_CHECKING, Generic, TypeVar
from weakref import ref as _ref

from .recursive_repr import recursive_repr

if TYPE_CHECKING:
    from typing import Any, Optional, Tuple, Type

__all__ = ["WeakReference"]


T = TypeVar("T")  # Any type.


class WeakReference(Generic[T], object):
    """
    Weak reference object that supports pickling.

    Inherits from:
      - :class:`typing.Generic`

    .. code:: python

        >>> from objetto.utils.weak_reference import WeakReference

        >>> class MyClass(object):
        ...     pass
        ...
        >>> strong = MyClass()
        >>> weak = WeakReference(strong)
        >>> weak() is strong
        True
        >>> del strong
        >>> weak() is None
        True

    :param obj: Object to reference.
    :type obj: object
    """

    __slots__ = ("__weakref__", "__ref")

    def __init__(self, obj=None):
        # type: (T) -> None
        if obj is None:
            self.__ref = _ref(
                type("Dead", (object,), {"__slots__": ("__weakref__",)})()
            )
        else:
            self.__ref = _ref(obj)

    def __hash__(self):
        """
        Get hash.

        :return: Hash.
        :rtype: int
        """
        return hash(self.__ref)

    def __eq__(self, other):
        # type: (Any) -> bool
        """
        Compare for equality.

        :param other: Other.

        :return: True if equal.
        :rtype: bool or NotImplemented
        """
        if self is other:
            return True
        if not isinstance(other, WeakReference):
            return False
        return self.__ref == other.__ref

    def __ne__(self, other):
        # type: (Any) -> bool
        """
        Compare for inequality.

        :param other: Other.

        :return: True if not equal.
        :rtype: bool or NotImplemented
        """
        result = self.__eq__(other)
        if result is NotImplemented:
            return NotImplemented
        return not result

    @recursive_repr
    def __repr__(self):
        # type: () -> str
        """
        Get representation.

        :return: Representation.
        :rtype: str
        """
        obj = self()
        return "<{} object at {}; {}>".format(
            type(self).__name__,
            hex(id(self)),
            "dead"
            if obj is None
            else "to '{}' at {}".format(
                type(obj).__name__,
                hex(id(obj)),
            ),
        )

    def __str__(self):
        # type: () -> str
        """
        Get string representation.

        :return: String representation.
        :rtype: str
        """
        return self.__repr__()

    def __call__(self):
        # type: () -> Optional[T]
        """
        Get strong reference to the object or None if no longer alive.

        :return: Strong reference or `None`.
        :rtype: object or None
        """
        return self.__ref()

    def __reduce__(self):
        # type: () -> Tuple[Type[WeakReference], Tuple[Optional[T]]]
        """
        Reduce for pickling.

        :return: Class and strong reference to object.
        :rtype: tuple[type[objetto.utils.weak_reference.WeakReference], tuple[object]]
        """
        return type(self), (self(),)



# -*- coding: utf-8 -*-
"""Object changes."""

from typing import TYPE_CHECKING, cast

from six import integer_types, string_types

from ._bases import final
from ._states import BaseState
from .data import (
    Data,
    data_attribute,
    data_dict_attribute,
    data_list_attribute,
    data_set_attribute,
)

if TYPE_CHECKING:
    from typing import Any, Callable, Final, Optional

    from ._data import DataAttribute, DictData, ListData, SetData
    from ._history import HistoryObject
    from ._objects import BaseObject

__all__ = [
    "BaseChange",
    "BaseAtomicChange",
    "Batch",
    "Update",
    "DictUpdate",
    "ListInsert",
    "ListDelete",
    "ListUpdate",
    "ListMove",
    "SetUpdate",
    "SetRemove",
]


class BaseChange(Data):
    """
    Base change.

    Inherits from:
      - :class:`objetto.data.Data`

    Inherited By:
      - :class:`objetto.bases.BaseAtomicChange`
      - :class:`objetto.changes.Batch`
    """

    name = data_attribute(
        string_types, checked=False, abstracted=True
    )  # type: DataAttribute[str]
    """
    Name describing the change.

    :type: str
    """

    obj = data_attribute(
        ".._objects|BaseObject", subtypes=True, checked=False, finalized=True
    )  # type: Final[DataAttribute[BaseObject]]
    """
    Object being changed.

    :type: objetto.bases.BaseObject
    """


class BaseAtomicChange(BaseChange):
    """
    Base atomic change.

    Inherits from:
      - :class:`objetto.bases.BaseChange`

    Inherited By:
      - :class:`objetto.changes.Update`
      - :class:`objetto.changes.DictUpdate`
      - :class:`objetto.changes.ListInsert`
      - :class:`objetto.changes.ListDelete`
      - :class:`objetto.changes.ListUpdate`
      - :class:`objetto.changes.ListMove`
      - :class:`objetto.changes.SetUpdate`
      - :class:`objetto.changes.SetRemove`
    """

    __redo__ = data_attribute(
        finalized=True, compared=False, serialized=False, represented=False
    )  # type: Final[DataAttribute[Callable]]
    """Redo delegate."""

    __undo__ = data_attribute(
        finalized=True, compared=False, serialized=False, represented=False
    )  # type: Final[DataAttribute[Callable]]
    """Undo delegate."""

    old_state = data_attribute(
        BaseState, subtypes=True, checked=False, finalized=True
    )  # type: Final[DataAttribute[BaseState]]
    """
    Object state before the change.

    :type: objetto.bases.BaseState
    """

    new_state = data_attribute(
        BaseState, subtypes=True, checked=False, finalized=True
    )  # type: Final[DataAttribute[BaseState]]
    """
    Object state after the change.

    :type: objetto.bases.BaseState
    """

    old_children = cast(
        "DataAttribute[SetData[BaseObject]]",
        data_set_attribute(
            ".._objects|BaseObject",
            subtypes=True,
            checked=False,
            finalized=True,
            interactive=False,
        ),
    )  # type: Final[DataAttribute[SetData[BaseObject]]]
    """
    Children objects being released.

    :type: objetto.data.SetData[objetto.bases.BaseObject]
    """

    new_children = cast(
        "DataAttribute[SetData[BaseObject]]",
        data_set_attribute(
            ".._objects|BaseObject",
            subtypes=True,
            checked=False,
            finalized=True,
            interactive=False,
        ),
    )  # type: Final[DataAttribute[SetData[BaseObject]]]
    """
    Children objects being adopted.

    :type: objetto.data.SetData[objetto.bases.BaseObject]
    """

    history_adopters = cast(
        "DataAttribute[SetData[BaseObject]]",
        data_set_attribute(
            ".._objects|BaseObject",
            subtypes=True,
            checked=False,
            finalized=True,
            interactive=False,
        ),
    )  # type: Final[DataAttribute[SetData[BaseObject]]]
    """
    Objects adopting the history from the object being changed.

    :type: objetto.data.SetData[objetto.bases.BaseObject]
    """

    history = data_attribute(
        (".._history|HistoryObject", None),
        subtypes=False,
        checked=False,
        finalized=True,
        default=None,
    )  # type: Final[DataAttribute[Optional[HistoryObject]]]
    """
    History where this changed originated from (result of an redo/undo operation).

    :type: objetto.history.HistoryObject or None
    """


@final
class Batch(BaseChange):
    """
    Batch change.

    Inherits from:
      - :class:`objetto.bases.BaseChange`
    """

    name = data_attribute(string_types, checked=False)  # type: DataAttribute[str]
    """
    Name describing the change.

    :type: str
    """

    metadata = cast(
        "DataAttribute[DictData[str, Any]]",
        data_dict_attribute(key_types=string_types, checked=False, interactive=False),
    )  # type: DataAttribute[DictData[str, Any]]
    """
    Metadata.

    :type: objetto.data.DictData[str, Any]
    """


@final
class Update(BaseAtomicChange):
    """
    Object's attributes have been updated.

    Inherits from:
      - :class:`objetto.bases.BaseAtomicChange`
    """

    name = data_attribute(
        string_types, checked=False, default="Update Attributes"
    )  # type: DataAttribute[str]
    """
    Name describing the change.

    :type: str
    """

    old_values = cast(
        "DataAttribute[DictData[str, Any]]",
        data_dict_attribute(checked=False, interactive=False),
    )  # type: DataAttribute[DictData[str, Any]]
    """
    Old attribute values.

    :type: objetto.data.DictData[str, Any]
    """

    new_values = cast(
        "DataAttribute[DictData[str, Any]]",
        data_dict_attribute(checked=False, interactive=False),
    )  # type: DataAttribute[DictData[str, Any]]
    """
    New attribute values.

    :type: objetto.data.DictData[str, Any]
    """


@final
class DictUpdate(BaseAtomicChange):
    """
    Dictionary values have been updated.

    Inherits from:
      - :class:`objetto.bases.BaseAtomicChange`
    """

    name = data_attribute(
        string_types, checked=False, default="Update Values"
    )  # type: DataAttribute[str]
    """
    Name describing the change.
    """

    old_values = cast(
        "DataAttribute[DictData[Any, Any]]",
        data_dict_attribute(checked=False, interactive=False),
    )  # type: DataAttribute[DictData[Any, Any]]
    """
    Old values.

    :type: objetto.data.DictData[collections.abc.Hashable, Any]
    """

    new_values = cast(
        "DataAttribute[DictData[Any, Any]]",
        data_dict_attribute(checked=False, interactive=False),
    )  # type: DataAttribute[DictData[Any, Any]]
    """
    New values.

    :type: objetto.data.DictData[collections.abc.Hashable, Any]
    """


@final
class ListInsert(BaseAtomicChange):
    """
    Values have been inserted into the list.

    Inherits from:
      - :class:`objetto.bases.BaseAtomicChange`
    """

    name = data_attribute(
        string_types, checked=False, default="Insert Values"
    )  # type: DataAttribute[str]
    """
    Name describing the change.

    :type: str
    """

    index = data_attribute(integer_types, checked=False)  # type: DataAttribute[int]
    """
    Insertion index.

    :type: int
    """

    last_index = data_attribute(
        integer_types, checked=False
    )  # type: DataAttribute[int]
    """
    Last inserted value index.

    :type: int
    """

    stop = data_attribute(integer_types, checked=False)  # type: DataAttribute[int]
    """
    Stop index.

    :type: int
    """

    new_values = cast(
        "DataAttribute[ListData[Any]]",
        data_list_attribute(checked=False, interactive=False),
    )  # type: DataAttribute[ListData[Any]]
    """
    New values.

    :type: objetto.data.ListData[Any]
    """


@final
class ListDelete(BaseAtomicChange):
    """
    Values have been removed from the list.

    Inherits from:
      - :class:`objetto.bases.BaseAtomicChange`
    """

    name = data_attribute(
        string_types, checked=False, default="Remove Values"
    )  # type: DataAttribute[str]
    """
    Name describing the change.

    :type: str
    """

    index = data_attribute(integer_types, checked=False)  # type: DataAttribute[int]
    """
    First removed value index.

    :type: int
    """

    last_index = data_attribute(
        integer_types, checked=False
    )  # type: DataAttribute[int]
    """
    Last removed value index.

    :type: int
    """

    stop = data_attribute(integer_types, checked=False)  # type: DataAttribute[int]
    """
    Stop index.

    :type: int
    """

    old_values = cast(
        "DataAttribute[ListData[Any]]",
        data_list_attribute(checked=False, interactive=False),
    )  # type: DataAttribute[ListData[Any]]
    """
    Old values.

    :type: objetto.data.ListData[Any]
    """


@final
class ListUpdate(BaseAtomicChange):
    """
    List values have been updated.

    Inherits from:
      - :class:`objetto.bases.BaseAtomicChange`
    """

    name = data_attribute(
        string_types, checked=False, default="Update values"
    )  # type: DataAttribute[str]
    """
    Name describing the change.

    :type: str
    """

    index = data_attribute(integer_types, checked=False)  # type: DataAttribute[int]
    """
    First updated value index.

    :type: int
    """

    last_index = data_attribute(
        integer_types, checked=False
    )  # type: DataAttribute[int]
    """
    Last updated value index.

    :type: int
    """

    stop = data_attribute(integer_types, checked=False)  # type: DataAttribute[int]
    """
    Stop index.

    :type: int
    """

    old_values = cast(
        "DataAttribute[ListData[Any]]",
        data_list_attribute(checked=False, interactive=False),
    )  # type: DataAttribute[ListData[Any]]
    """
    Old values.

    :type: objetto.data.ListData[Any]
    """

    new_values = cast(
        "DataAttribute[ListData[Any]]",
        data_list_attribute(checked=False, interactive=False),
    )  # type: DataAttribute[ListData[Any]]
    """
    New values.

    :type: objetto.data.ListData[Any]
    """


@final
class ListMove(BaseAtomicChange):
    """
    List values have been moved internally.

    Inherits from:
      - :class:`objetto.bases.BaseAtomicChange`
    """

    name = data_attribute(
        string_types, checked=False, default="Move values"
    )  # type: DataAttribute[str]
    """
    Name describing the change.

    :type: str
    """

    index = data_attribute(integer_types, checked=False)  # type: DataAttribute[int]
    """
    First moved value index.

    :type: int
    """

    last_index = data_attribute(
        integer_types, checked=False
    )  # type: DataAttribute[int]
    """
    Last moved value index.

    :type: int
    """

    stop = data_attribute(integer_types, checked=False)  # type: DataAttribute[int]
    """
    Stop index.

    :type: int
    """

    target_index = data_attribute(
        integer_types, checked=False
    )  # type: DataAttribute[int]
    """
    Index where values are being moved to.

    :type: int
    """

    post_index = data_attribute(
        integer_types, checked=False
    )  # type: DataAttribute[int]
    """
    First moved value index after the move.

    :type: int
    """

    post_last_index = data_attribute(
        integer_types, checked=False
    )  # type: DataAttribute[int]
    """
    Last moved value index after the move.

    :type: int
    """

    post_stop = data_attribute(integer_types, checked=False)  # type: DataAttribute[int]
    """
    Stop index after the move.

    :type: int
    """

    values = cast(
        "DataAttribute[ListData[Any]]",
        data_list_attribute(checked=False, interactive=False),
    )  # type: DataAttribute[ListData[Any]]
    """
    Values being moved.

    :type: objetto.data.ListData[Any]
    """


@final
class SetUpdate(BaseAtomicChange):
    """
    Values have been added to the set.

    Inherits from:
      - :class:`objetto.bases.BaseAtomicChange`
    """

    name = data_attribute(
        string_types, checked=False, default="Add values"
    )  # type: DataAttribute[str]
    """
    Name describing the change.

    :type: str
    """

    new_values = cast(
        "DataAttribute[SetData[Any]]",
        data_set_attribute(checked=False, interactive=False),
    )  # type: DataAttribute[SetData[Any]]
    """
    Values being added to the set.

    :type: objetto.data.SetData[collections.abc.Hashable]
    """


@final
class SetRemove(BaseAtomicChange):
    """
    Values have been removed from the set.

    Inherits from:
      - :class:`objetto.bases.BaseAtomicChange`
    """

    name = data_attribute(
        string_types, checked=False, default="Remove values"
    )  # type: DataAttribute[str]
    """
    Name describing the change.

    :type: str
    """

    old_values = cast(
        "DataAttribute[SetData[Any]]",
        data_set_attribute(checked=False, interactive=False),
    )  # type: DataAttribute[SetData[Any]]
    """
    Values being removed from the set.

    :type: objetto.data.SetData[collections.abc.Hashable]
    """



# -*- coding: utf-8 -*-
"""Custom representation functions."""

from typing import TYPE_CHECKING

from six import iteritems

if TYPE_CHECKING:
    from typing import Any, Callable, Hashable, Iterable, Mapping, Optional, Tuple

__all__ = ["custom_mapping_repr", "custom_iterable_repr"]


def custom_mapping_repr(
    mapping,  # type: Mapping
    prefix="{",  # type: str
    template="{key}: {value}",  # type: str
    separator=", ",  # type: str
    suffix="}",  # type: str
    sorting=False,  # type: bool
    sort_key=None,  # type: Optional[Callable[[Any], Any]]
    reverse=False,  # type: bool
    key_repr=repr,  # type: Callable[[Any], Any]
    value_repr=repr,  # type: Callable[[Any], Any]
):
    # type: (...) -> str
    """
    Get custom representation of a mapping.

    .. code:: python

        >>> from objetto.utils.custom_repr import custom_mapping_repr

        >>> dct = {"a": 1, "b": 2}
        >>> custom_mapping_repr(
        ...     dct, prefix="<", suffix=">", template="{key}={value}", sorting=True
        ... )
        "<'a'=1, 'b'=2>"

    :param mapping: Mapping.
    :type mapping: collections.abc.Mapping

    :param prefix: Prefix.
    :type prefix: str

    :param template: Item format template ({key} and {value}).
    :type template: str

    :param separator: Separator.
    :type separator: str

    :param suffix: Suffix.
    :type suffix: str

    :param sorting: Whether to sort keys.
    :type sorting: bool

    :param sort_key: Sorting key.
    :type sort_key: function or None

    :param reverse: Reverse sorting.
    :type reverse: bool

    :param key_repr: Key representation function.
    :type key_repr: function

    :param value_repr: Value representation function.
    :type value_repr: function

    :return: Custom representation.
    :rtype: str
    """
    parts = []
    iterable = iteritems(mapping)  # type: Iterable[Tuple[Hashable, Any]]
    if sort_key is None:
        sort_key = lambda item: item[0]
    if sorting:
        iterable = sorted(iterable, key=sort_key, reverse=reverse)
    for key, value in iterable:
        part = template.format(key=key_repr(key), value=value_repr(value))
        parts.append(part)
    return prefix + separator.join(parts) + suffix


def custom_iterable_repr(
    iterable,  # type: Iterable
    prefix="[",  # type: str
    template="{value}",  # type: str
    separator=", ",  # type: str
    suffix="]",  # type: str
    sorting=False,  # type: bool
    sort_key=None,  # type: Optional[Callable[[Any], Any]]
    reverse=False,  # type: bool
    value_repr=repr,  # type: Callable[[Any], Any]
):
    # type: (...) -> str
    """
    Get custom representation of an iterable.

    .. code:: python

        >>> from objetto.utils.custom_repr import custom_iterable_repr

        >>> tup = ("a", "b", "c", 1, 2, 3)
        >>> custom_iterable_repr(tup, prefix="<", suffix=">", value_repr=str)
        '<a, b, c, 1, 2, 3>'

    :param iterable: Iterable.
    :type iterable: collections.abc.Iterable

    :param prefix: Prefix.
    :type prefix: str

    :param template: Item format template ({key} and {value}).
    :type template: str

    :param separator: Separator.
    :type separator: str

    :param suffix: Suffix.
    :type suffix: str

    :param sorting: Whether to sort the iterable or not.
    :type sorting: bool

    :param sort_key: Sorting key.
    :type sort_key: function

    :param reverse: Reverse sorting.
    :type reverse: bool

    :param value_repr: Value representation function.
    :type value_repr: function

    :return: Custom representation.
    :rtype: str
    """
    parts = []
    if sorting:
        iterable = sorted(iterable, key=sort_key, reverse=reverse)
    for value in iterable:
        part = template.format(value=value_repr(value))
        parts.append(part)
    return prefix + separator.join(parts) + suffix



# -*- coding: utf-8 -*-
"""Immutable set state."""

from typing import TYPE_CHECKING, TypeVar, cast

try:
    import collections.abc as collections_abc
except ImportError:
    import collections as collections_abc  # type: ignore

from pyrsistent import pset
from six import iteritems

from .._bases import BaseInteractiveSet, final
from ..utils.custom_repr import custom_iterable_repr, custom_mapping_repr
from ..utils.recursive_repr import recursive_repr
from .bases import BaseState

if TYPE_CHECKING:
    from typing import Any, Iterable, Iterator, Type

    from pyrsistent.typing import PSet

__all__ = ["SetState"]


T = TypeVar("T")  # Any type.


# noinspection PyTypeChecker
_SS = TypeVar("_SS", bound="SetState")


@final
class SetState(BaseState[T], BaseInteractiveSet[T]):
    """
    Immutable set state.

    Inherits from:
      - :class:`objetto.bases.BaseState`
      - :class:`objetto.bases.BaseInteractiveSet`

    :param initial: Initial values.
    :type initial: collections.abc.Iterable
    """

    __slots__ = ()

    @classmethod
    def _make(cls, internal=pset()):
        # type: (Type[_SS], PSet[T]) -> _SS
        """
        Make new state by directly setting the internal state.

        :param internal: Internal state.
        :return: State.
        """
        return super(SetState, cls)._make(internal)

    @staticmethod
    def _make_internal(initial):
        # type: (Iterable[T]) -> PSet[T]
        """
        Initialize internal state.

        :param initial: Initial values.
        """
        return pset(initial)

    @classmethod
    def _from_iterable(cls, iterable):
        # type: (Iterable) -> SetState
        """
        Make set state from iterable.

        :param iterable: Iterable.
        :type iterable: collections.abc.Iterable

        :return: Set state.
        :rtype: objetto.states.SetState
        """
        if isinstance(iterable, type(pset())):
            return SetState._make(iterable)
        else:
            return SetState(iterable)

    def __init__(self, initial=()):
        # type: (Iterable[T]) -> None
        super(SetState, self).__init__(initial=initial)

    def __hash__(self):
        # type: () -> int
        """
        Get hash.

        :return: Hash.
        :rtype: int
        """
        return super(SetState, self).__hash__()

    def __eq__(self, other):
        # type: (object) -> bool
        """
        Compare for equality.

        :param other: Another object.

        :return: True if equal.
        :rtype: bool
        """
        if self is other:
            return True
        if not isinstance(other, collections_abc.Hashable):
            return self._internal == other
        if isinstance(other, SetState):
            return self._internal == other._internal
        return False

    def __contains__(self, value):
        # type: (Any) -> bool
        """
        Get whether value is present.

        :param value: Value.
        :type value: collections.abc.Hashable

        :return: True if contains.
        :rtype: bool
        """
        return value in self._internal

    def __iter__(self):
        # type: () -> Iterator[T]
        """
        Iterate over values.

        :return: Values iterator.
        :rtype: collections.abc.Iterator[collections.abc.Hashable]
        """
        for key in self._internal:
            yield key

    def __len__(self):
        # type: () -> int
        """
        Get value count.

        :return: Value count.
        :rtype: int
        """
        return len(self._internal)

    @recursive_repr
    def __repr__(self):
        # type: () -> str
        """
        Get representation.

        :return: Representation.
        :rtype: str
        """
        return custom_iterable_repr(
            self._internal,
            prefix="{}([".format(type(self).__name__),
            suffix="])",
            sorting=True,
            sort_key=lambda v: hash(v),
        )

    def _hash(self):
        # type: () -> int
        """
        Get hash.

        :return: Hash.
        :rtype: int
        """
        return hash(self)

    def _clear(self):
        # type: (_SS) -> _SS
        """
        Clear.

        :return: Transformed.
        :rtype: objetto.states.SetState
        """
        return self._make()

    def _add(self, value):
        # type: (_SS, T) -> _SS
        """
        Add value.

        :param value: Value.
        :type value: collections.abc.Hashable

        :return: Transformed.
        :rtype: objetto.states.SetState
        """
        return self._make(self._internal.add(value))

    def _discard(self, *values):
        # type: (_SS, T) -> _SS
        """
        Discard value(s).

        :param values: Value(s).
        :type values: collections.abc.Hashable

        :return: Transformed.
        :rtype: objetto.states.SetState

        :raises ValueError: No values provided.
        """
        if not values:
            error = "no values provided"
            raise ValueError(error)
        return self._make(self._internal.discard(*values))

    def _remove(self, *values):
        # type: (_SS, T) -> _SS
        """
        Remove existing value(s).

        :param values: Value(s).
        :type values: collections.abc.Hashable

        :return: Transformed.
        :rtype: objetto.states.SetState

        :raises ValueError: No values provided.
        :raises KeyError: Value is not present.
        """
        if not values:
            error = "no values provided"
            raise ValueError(error)
        return self._make(self._internal.difference(values))

    def _replace(self, old_value, new_value):
        # type: (_SS, T, T) -> _SS
        """
        Replace existing value with a new one.

        :param old_value: Existing value.
        :type old_value: collections.abc.Hashable

        :param new_value: New value.
        :type new_value: collections.abc.Hashable

        :return: Transformed.
        :rtype: objetto.states.SetState

        :raises KeyError: Value is not present.
        """
        return self._make(self._internal.remove(old_value).add(new_value))

    def _update(self, iterable):
        # type: (_SS, Iterable[T]) -> _SS
        """
        Update with iterable.

        :param iterable: Iterable.
        :type iterable: collections.abc.Iterable[collections.abc.Hashable]

        :return: Transformed.
        :rtype: objetto.states.SetState
        """
        return self._make(self._internal.update(iterable))

    def isdisjoint(self, iterable):
        # type: (Iterable) -> bool
        """
        Get whether is a disjoint set of an iterable.

        :param iterable: Iterable.
        :type iterable: collections.abc.Iterable

        :return: True if is disjoint.
        :rtype: bool
        """
        return self._internal.isdisjoint(iterable)

    def issubset(self, iterable):
        # type: (Iterable) -> bool
        """
        Get whether is a subset of an iterable.

        :param iterable: Iterable.
        :type iterable: collections.abc.Iterable

        :return: True if is subset.
        :rtype: bool
        """
        return self._internal.issubset(iterable)

    def issuperset(self, iterable):
        # type: (Iterable) -> bool
        """
        Get whether is a superset of an iterable.

        :param iterable: Iterable.
        :type iterable: collections.abc.Iterable

        :return: True if is superset.
        :rtype: bool
        """
        return self._internal.issuperset(iterable)

    def intersection(self, iterable):
        # type: (Iterable) -> SetState
        """
        Get intersection.

        :param iterable: Iterable.
        :type iterable: collections.abc.Iterable

        :return: Intersection.
        :rtype: objetto.states.SetState
        """
        return SetState._make(self._internal.intersection(iterable))

    def difference(self, iterable):
        # type: (Iterable) -> SetState
        """
        Get difference.

        :param iterable: Iterable.
        :type iterable: collections.abc.Iterable

        :return: Difference.
        :rtype: objetto.states.SetState
        """
        return SetState._make(self._internal.difference(iterable))

    def inverse_difference(self, iterable):
        # type: (Iterable) -> SetState
        """
        Get an iterable's difference to this.

        :param iterable: Iterable.
        :type iterable: collections.abc.Iterable

        :return: Inverse Difference.
        :rtype: objetto.states.SetState
        """
        return SetState._make(pset(iterable).difference(self._internal))

    def symmetric_difference(self, iterable):
        # type: (Iterable) -> SetState
        """
        Get symmetric difference.

        :param iterable: Iterable.
        :type iterable: collections.abc.Iterable

        :return: Symmetric difference.
        :rtype: objetto.states.SetState
        """
        return SetState._make(self._internal.symmetric_difference(iterable))

    def union(self, iterable):
        # type: (Iterable) -> SetState
        """
        Get union.

        :param iterable: Iterable.
        :type iterable: collections.abc.Iterable

        :return: Union.
        :rtype: objetto.states.SetState
        """
        return SetState._make(self._internal.union(iterable))

    def find_with_attributes(self, **attributes):
        # type: (Any) -> T
        """
        Find first value that matches unique attribute values.

        :param attributes: Attributes to match.

        :return: Value that has matching attributes.

        :raises ValueError: No attributes provided or no match found.
        """
        if not attributes:
            error = "no attributes provided"
            raise ValueError(error)
        for value in self._internal:
            for a_name, a_value in iteritems(attributes):
                if not hasattr(value, a_name) or getattr(value, a_name) != a_value:
                    break
            else:
                return value
        error = "could not find a match for {}".format(
            custom_mapping_repr(
                attributes,
                prefix="(",
                template="{key}={value}",
                suffix=")",
                key_repr=str,
            ),
        )
        raise ValueError(error)

    @property
    def _internal(self):
        # type: () -> PSet[T]
        """Internal values."""
        return cast("PSet[T]", super(SetState, self)._internal)



# -*- coding: utf-8 -*-
"""List data structures."""

from typing import TYPE_CHECKING, TypeVar, cast, overload

try:
    import collections.abc as collections_abc
except ImportError:
    import collections as collections_abc  # type: ignore

from six import with_metaclass

from .._bases import final
from .._states import BaseState, ListState
from .._structures import (
    BaseInteractiveListStructure,
    BaseListStructure,
    BaseListStructureMeta,
    SerializationError,
)
from .bases import (
    BaseAuxiliaryData,
    BaseAuxiliaryDataMeta,
    BaseInteractiveAuxiliaryData,
)

if TYPE_CHECKING:
    from typing import Any, Iterable, List, Sequence, Type, Union


__all__ = ["ListData", "InteractiveListData"]


T = TypeVar("T")  # Any type.


class ListDataMeta(BaseAuxiliaryDataMeta, BaseListStructureMeta):
    """Metaclass for `ListData`."""

    @property
    @final
    def _base_auxiliary_type(cls):
        # type: () -> Type[ListData]
        """Base auxiliary container type."""
        return ListData


# noinspection PyTypeChecker
_LD = TypeVar("_LD", bound="ListData")


class ListData(
    with_metaclass(
        ListDataMeta,
        BaseListStructure[T],
        BaseAuxiliaryData[T],
    )
):
    """
    List data.

    Inherits from:
      - :class:`objetto.bases.BaseListStructure`
      - :class:`objetto.bases.BaseAuxiliaryData`

    Inherited by:
      - :class:`objetto.data.InteractiveListData`

    :param initial: Initial values.
    :type initial: collections.abc.Iterable
    """

    __slots__ = ()

    @classmethod
    @final
    def __make__(cls, state=ListState()):
        # type: (Type[_LD], BaseState) -> _LD
        """
        Make a new list data.

        :param state: Internal state.
        :return: New list data.
        """
        return super(ListData, cls).__make__(state)

    @final
    def __init__(self, initial=()):
        # type: (Iterable[T]) -> None
        self._init_state(self.__get_initial_state(initial))

    @overload
    def __getitem__(self, index):
        # type: (int) -> T
        pass

    @overload
    def __getitem__(self, index):
        # type: (slice) -> Sequence[T]
        pass

    @final
    def __getitem__(self, index):
        """
        Get value/values at index/from slice.

        :param index: Index/slice.
        :type index: int or slice

        :return: Value/values.
        :rtype: Any or objetto.states.ListState
        """
        return self._state[index]

    @classmethod
    @final
    def __get_initial_state(cls, input_values, factory=True):
        # type: (Iterable[T], bool) -> ListState[T]
        """
        Get initial state.

        :param input_values: Input values.
        :param factory: Whether to run values through factory.
        :return: Initial state.
        """
        if not cls._relationship.passthrough:
            state = ListState(
                cls._relationship.fabricate_value(v, factory=factory)
                for v in input_values
            )
        else:
            state = ListState(input_values)
        return state

    @final
    def _clear(self):
        # type: (_LD) -> _LD
        """
        Clear all values.

        :return: Transformed.
        :rtype: objetto.data.ListData
        """
        return type(self).__make__()

    @final
    def _insert(self, index, *values):
        # type: (_LD, int, T) -> _LD
        """
        Insert value(s) at index.

        :param index: Index.
        :type index: int

        :param values: Value(s).
        :type values: collections.abc.Iterable

        :return: Transformed.
        :rtype: objetto.data.ListData

        :raises ValueError: No values provided.
        """
        cls = type(self)
        if not cls._relationship.passthrough:
            fabricated_values = (cls._relationship.fabricate_value(v) for v in values)
            return type(self).__make__(self._state.insert(index, *fabricated_values))
        else:
            return type(self).__make__(self._state.insert(index, *values))

    @final
    def _append(self, value):
        # type: (_LD, T) -> _LD
        """
        Append value at the end.

        :param value: Value.

        :return: Transformed.
        :rtype: objetto.data.ListData
        """
        cls = type(self)
        fabricated_value = cls._relationship.fabricate_value(value)
        return type(self).__make__(self._state.append(fabricated_value))

    @final
    def _extend(self, iterable):
        # type: (_LD, Iterable[T]) -> _LD
        """
        Extend at the end with iterable.

        :param iterable: Iterable.
        :type iterable: collections.abc.Iterable

        :return: Transformed.
        :rtype: objetto.data.ListData
        """
        cls = type(self)
        if not cls._relationship.passthrough:
            fabricated_iterable = (
                cls._relationship.fabricate_value(v) for v in iterable
            )
            return type(self).__make__(self._state.extend(fabricated_iterable))
        else:
            return type(self).__make__(self._state.extend(iterable))

    @final
    def _remove(self, value):
        # type: (_LD, T) -> _LD
        """
        Remove first occurrence of value.

        :param value: Value.

        :return: Transformed.
        :rtype: objetto.data.ListData

        :raises ValueError: Value is not present.
        """
        return type(self).__make__(self._state.remove(value))

    @final
    def _reverse(self):
        # type: (_LD) -> _LD
        """
        Reverse values.

        :return: Transformed.
        :rtype: objetto.data.ListData
        """
        return type(self).__make__(self._state.reverse())

    @final
    def _move(self, item, target_index):
        # type: (_LD, Union[slice, int], int) -> _LD
        """
        Move values internally.

        :param item: Index/slice.
        :type item: int or slice

        :param target_index: Target index.
        :type target_index: int

        :return: Transformed.
        :rtype: objetto.data.ListData
        """
        return type(self).__make__(self._state.move(item, target_index))

    @final
    def _delete(self, item):
        # type: (_LD, Union[slice, int]) -> _LD
        """
        Delete values at index/slice.

        :param item: Index/slice.
        :type item: int or slice

        :return: Transformed.
        :rtype: objetto.data.ListData
        """
        return type(self).__make__(self._state.delete(item))

    @final
    def _update(self, index, *values):
        # type: (_LD, int, T) -> _LD
        """
        Update value(s) starting at index.

        :param index: Index.
        :type index: int

        :param values: Value(s).

        :return: Transformed.
        :rtype: objetto.data.ListData

        :raises ValueError: No values provided.
        """
        cls = type(self)
        if not cls._relationship.passthrough:
            fabricated_values = (cls._relationship.fabricate_value(v) for v in values)
            return type(self).__make__(self._state.update(index, *fabricated_values))
        else:
            return type(self).__make__(self._state.update(index, *values))

    @classmethod
    @final
    def deserialize(cls, serialized, **kwargs):
        # type: (Type[_LD], List, Any) -> _LD
        """
        Deserialize.

        :param serialized: Serialized.
        :type: serialized: list

        :param kwargs: Keyword arguments to be passed to the deserializers.

        :return: Deserialized.
        :rtype: objetto.data.ListData

        :raises objetto.exceptions.SerializationError: Can't deserialize.
        """
        if not cls._relationship.serialized:
            error = "'{}' is not deserializable".format(cls.__name__)
            raise SerializationError(error)
        state = ListState(
            cls.deserialize_value(v, location=None, **kwargs) for v in serialized
        )
        return cls.__make__(state)

    @final
    def serialize(self, **kwargs):
        # type: (Any) -> List
        """
        Serialize.

        :param kwargs: Keyword arguments to be passed to the serializers.

        :return: Serialized.
        :rtype: list

        :raises objetto.exceptions.SerializationError: Can't serialize.
        """
        if not type(self)._relationship.serialized:
            error = "'{}' is not serializable".format(type(self).__fullname__)
            raise SerializationError(error)
        return list(
            self.serialize_value(v, location=None, **kwargs) for v in self._state
        )

    @property
    @final
    def _state(self):
        # type: () -> ListState[T]
        """
        Internal state.

        :rtype: objetto.states.ListState
        """
        return cast("ListState", super(BaseListStructure, self)._state)


class InteractiveListData(
    ListData[T],
    BaseInteractiveListStructure[T],
    BaseInteractiveAuxiliaryData[T],
):
    """
    Interactive list data.

    Inherits from:
      - :class:`objetto.data.ListData`
      - :class:`objetto.bases.BaseInteractiveListStructure`
      - :class:`objetto.bases.BaseInteractiveAuxiliaryData`
    """

    __slots__ = ()



# -*- coding: utf-8 -*-
"""Base types."""

from ._applications import Phase
from ._bases import (
    Base,
    BaseCollection,
    BaseContainer,
    BaseDict,
    BaseHashable,
    BaseInteractiveCollection,
    BaseInteractiveDict,
    BaseInteractiveList,
    BaseInteractiveSet,
    BaseIterable,
    BaseList,
    BaseMutableCollection,
    BaseMutableDict,
    BaseMutableList,
    BaseMutableSet,
    BaseProtectedCollection,
    BaseProtectedDict,
    BaseProtectedList,
    BaseProtectedSet,
    BaseSet,
    BaseSized,
    abstract_member,
    final,
    init,
    init_context,
    make_base_cls,
    simplify_member_names,
)
from ._changes import BaseAtomicChange, BaseChange
from ._data import (
    BaseAuxiliaryData,
    BaseData,
    BaseInteractiveAuxiliaryData,
    BaseInteractiveData,
)
from ._exceptions import BaseObjettoException
from ._factories import BaseFactory
from ._objects import (
    BaseAuxiliaryObject,
    BaseMutableAuxiliaryObject,
    BaseMutableObject,
    BaseObject,
    BaseProxyObject,
    BaseReaction,
)
from ._states import BaseState
from ._structures import (
    BaseAttribute,
    BaseAttributeStructure,
    BaseAuxiliaryStructure,
    BaseDictStructure,
    BaseInteractiveAttributeStructure,
    BaseInteractiveAuxiliaryStructure,
    BaseInteractiveDictStructure,
    BaseInteractiveListStructure,
    BaseInteractiveSetStructure,
    BaseInteractiveStructure,
    BaseListStructure,
    BaseMutableAttributeStructure,
    BaseMutableAuxiliaryStructure,
    BaseMutableDictStructure,
    BaseMutableListStructure,
    BaseMutableSetStructure,
    BaseMutableStructure,
    BaseRelationship,
    BaseSetStructure,
    BaseStructure,
)

__all__ = [
    "final",
    "abstract_member",
    "init",
    "init_context",
    "simplify_member_names",
    "make_base_cls",
    "BaseObjettoException",
    "Base",
    "BaseCollection",
    "BaseContainer",
    "BaseDict",
    "BaseHashable",
    "BaseInteractiveCollection",
    "BaseInteractiveDict",
    "BaseInteractiveList",
    "BaseInteractiveSet",
    "BaseIterable",
    "BaseList",
    "BaseMutableCollection",
    "BaseMutableDict",
    "BaseMutableList",
    "BaseMutableSet",
    "BaseProtectedCollection",
    "BaseProtectedDict",
    "BaseProtectedList",
    "BaseProtectedSet",
    "BaseSet",
    "BaseSized",
    "BaseState",
    "BaseRelationship",
    "BaseStructure",
    "BaseInteractiveStructure",
    "BaseMutableStructure",
    "BaseAttribute",
    "BaseAttributeStructure",
    "BaseInteractiveAttributeStructure",
    "BaseMutableAttributeStructure",
    "BaseAuxiliaryStructure",
    "BaseInteractiveAuxiliaryStructure",
    "BaseMutableAuxiliaryStructure",
    "BaseDictStructure",
    "BaseInteractiveDictStructure",
    "BaseMutableDictStructure",
    "BaseListStructure",
    "BaseInteractiveListStructure",
    "BaseMutableListStructure",
    "BaseSetStructure",
    "BaseInteractiveSetStructure",
    "BaseMutableSetStructure",
    "BaseData",
    "BaseInteractiveData",
    "BaseAuxiliaryData",
    "BaseInteractiveAuxiliaryData",
    "BaseObject",
    "BaseReaction",
    "BaseMutableObject",
    "BaseAuxiliaryObject",
    "BaseMutableAuxiliaryObject",
    "BaseProxyObject",
    "BaseChange",
    "Phase",
    "BaseAtomicChange",
    "BaseFactory",
]



# -*- coding: utf-8 -*-
"""Immutable structures."""

from .bases import (
    BaseAuxiliaryData,
    BaseData,
    BaseInteractiveAuxiliaryData,
    BaseInteractiveData,
    DataRelationship,
)
from .data import Data, DataAttribute, InteractiveData
from .dict import DictData, InteractiveDictData
from .list import InteractiveListData, ListData
from .set import InteractiveSetData, SetData

__all__ = [
    "DataRelationship",
    "BaseData",
    "BaseInteractiveData",
    "DataAttribute",
    "Data",
    "InteractiveData",
    "BaseAuxiliaryData",
    "BaseInteractiveAuxiliaryData",
    "DictData",
    "InteractiveDictData",
    "ListData",
    "InteractiveListData",
    "SetData",
    "InteractiveSetData",
]



# -*- coding: utf-8 -*-
"""Set objects and proxy."""

from collections import Counter as ValueCounter
from typing import TYPE_CHECKING, TypeVar, cast

try:
    import collections.abc as collections_abc
except ImportError:
    import collections as collections_abc  # type: ignore

from six import with_metaclass

from .._applications import Application
from .._bases import FINAL_METHOD_TAG, BaseMutableSet, final, init_context
from .._changes import SetRemove, SetUpdate
from .._data import BaseData, InteractiveDictData, SetData
from .._states import DictState, SetState
from .._structures import (
    BaseMutableSetStructure,
    BaseSetStructure,
    BaseSetStructureMeta,
    SerializationError,
)
from .bases import (
    BaseAuxiliaryObject,
    BaseAuxiliaryObjectFunctions,
    BaseAuxiliaryObjectMeta,
    BaseMutableAuxiliaryObject,
    BaseObject,
    BaseProxyObject,
)

if TYPE_CHECKING:
    from typing import (
        Any,
        Callable,
        Counter,
        Hashable,
        Iterable,
        List,
        Optional,
        Set,
        Type,
    )

    from .._applications import Store
    from .._history import HistoryObject

__all__ = ["SetObject", "MutableSetObject", "ProxySetObject"]


T = TypeVar("T")  # Any type.


DATA_MAP_METADATA_KEY = "data_map"
"""Data child locations metadata key."""


@final
class SetObjectFunctions(BaseAuxiliaryObjectFunctions):
    """Static functions for `SetObject`."""

    __slots__ = ()

    @staticmethod
    def replace_child_data(store, child, data_location, new_child_data):
        # type: (Store, BaseObject, Any, BaseData) -> Store
        """
        Replace child data.

        :param store: Object's store.
        :param child: Child getting their data replaced.
        :param data_location: Location of the existing child's data.
        :param new_child_data: New child's data.
        :return: Updated object's store.
        """
        original_data = store.data
        assert original_data is not None
        data = cast("SetData", original_data)._replace(data_location, new_child_data)
        metadata = store.metadata.set(
            DATA_MAP_METADATA_KEY,
            store.metadata.get(
                DATA_MAP_METADATA_KEY,
                DictState(),
            ).set(child, new_child_data),
        )
        return store.update({"data": data, "metadata": metadata})

    @staticmethod
    def update(
        obj,  # type: SetObject
        input_values,  # type: Iterable[Hashable]
        factory=True,  # type: bool
        history=None,  # type: Optional[HistoryObject]
    ):
        # type: (...) -> None
        cls = type(obj)
        relationship = cls._relationship
        with obj.app.__.write_context(obj) as (read, write):
            if not input_values:
                return

            # Get state, data, and data map.
            store = read()
            state = old_state = store.state  # type: SetState
            data = store.data  # type: SetData
            metadata = store.metadata  # type: InteractiveDictData
            data_map = metadata.get(
                DATA_MAP_METADATA_KEY, DictState()
            )  # type: DictState

            # Prepare change information.
            child_counter = ValueCounter()  # type: Counter[BaseObject]
            new_children = set()  # type: Set[BaseObject]
            history_adopters = set()  # type: Set[BaseObject]
            new_values = set()  # type: Set[Any]

            # For every input value.
            for value in set(input_values):

                # Fabricate new value.
                if factory:
                    value = relationship.fabricate_value(
                        value,
                        factory=factory,
                        kwargs={"app": obj.app},
                    )
                new_values.add(value)

                # Child relationship.
                if relationship.child:
                    same_app = obj._in_same_application(value)
                    child = cast("BaseObject", value) if same_app else None

                    # Update children counter and new children set.
                    if child is not None:
                        child_counter[child] += 1
                        new_children.add(child)

                        # Add history adopter.
                        if relationship.history and same_app:
                            history_adopters.add(child)

                    # Update data.
                    if relationship.data:
                        data_relationship = relationship.data_relationship
                        assert data_relationship is not None
                        if child is not None:
                            if type(child)._unique_descriptor is None:
                                error = (
                                    "'{}' objects (which class doesn't define a "
                                    "unique descriptor) can't be added as a child of "
                                    "'{}' object (which is a set object and defines a "
                                    "data relationship)"
                                ).format(
                                    type(child).__fullname__,
                                    type(obj).__fullname__,
                                )
                                raise TypeError(error)

                            with child.app.__.write_context(value) as (v_read, _):
                                data_value = data_relationship.fabricate_value(
                                    v_read().data,
                                )
                        else:
                            data_value = data_relationship.fabricate_value(value)
                        data = data._add(data_value)
                        data_map = data_map.set(value, data_value)

                # Update state.
                state = state.add(value)

            # Store data_map in the metadata.
            metadata = metadata.set(DATA_MAP_METADATA_KEY, data_map)

            # Prepare change.
            change = SetUpdate(
                __redo__=SetObjectFunctions.redo_update,
                __undo__=SetObjectFunctions.undo_update,
                obj=obj,
                old_children=(),
                new_children=new_children,
                new_values=new_values,
                history_adopters=history_adopters,
                old_state=old_state,
                new_state=state,
                history=history,
            )
            write(state, data, metadata, child_counter, change)

    @staticmethod
    def redo_update(change):
        # type: (SetUpdate) -> None
        SetObjectFunctions.update(
            cast("SetObject", change.obj),
            change.new_values,
            factory=False,
            history=change.obj._history,
        )

    @staticmethod
    def undo_update(change):
        # type: (SetUpdate) -> None
        SetObjectFunctions.remove(
            cast("SetObject", change.obj),
            change.new_values,
            history=change.obj._history,
        )

    @staticmethod
    def remove(
        obj,  # type: SetObject
        input_values,  # type: Iterable[Hashable]
        history=None,  # type: Optional[HistoryObject]
    ):
        # type: (...) -> None
        cls = type(obj)
        relationship = cls._relationship
        with obj.app.__.write_context(obj) as (read, write):
            if not input_values:
                return

            # Get state, data, and data map.
            store = read()
            state = old_state = store.state  # type: SetState
            data = store.data  # type: SetData
            metadata = store.metadata  # type: InteractiveDictData
            data_map = metadata.get(
                DATA_MAP_METADATA_KEY, DictState()
            )  # type: DictState

            # Prepare change information.
            child_counter = ValueCounter()  # type: Counter[BaseObject]
            old_children = set()  # type: Set[BaseObject]
            old_values = set()  # type: Set[Hashable]

            # For every input value.
            for value in set(input_values):

                # Check if value is in the set.
                if value not in store.state:
                    raise KeyError(value)
                old_values.add(value)

                # Child relationship.
                if relationship.child:
                    same_app = obj._in_same_application(value)
                    child = cast("BaseObject", value) if same_app else None

                    # Update children counter and new children set.
                    if child is not None:
                        child_counter[child] -= 1
                        old_children.add(child)

                    # Update data.
                    if relationship.data:
                        data = data._remove(data_map[value])
                        data_map = data_map.remove(value)

                # Update state.
                state = state.remove(value)

            # Store data_map in the metadata.
            metadata = metadata.set(DATA_MAP_METADATA_KEY, data_map)

            # Prepare change.
            change = SetRemove(
                __redo__=SetObjectFunctions.redo_remove,
                __undo__=SetObjectFunctions.undo_remove,
                obj=obj,
                old_children=old_children,
                new_children=(),
                old_values=old_values,
                history_adopters=(),
                old_state=old_state,
                new_state=state,
                history=history,
            )
            write(state, data, metadata, child_counter, change)

    @staticmethod
    def redo_remove(change):
        # type: (SetRemove) -> None
        SetObjectFunctions.remove(
            cast("SetObject", change.obj),
            change.old_values,
            history=change.obj._history,
        )

    @staticmethod
    def undo_remove(change):
        # type: (SetRemove) -> None
        SetObjectFunctions.update(
            cast("SetObject", change.obj),
            change.old_values,
            factory=False,
            history=change.obj._history,
        )


type.__setattr__(cast(type, SetObjectFunctions), FINAL_METHOD_TAG, True)


class SetObjectMeta(BaseAuxiliaryObjectMeta, BaseSetStructureMeta):
    """Metaclass for `SetObject`."""

    @property
    @final
    def _state_factory(cls):
        # type: () -> Callable[..., SetState]
        """State factory."""
        return SetState

    @property
    @final
    def _base_auxiliary_type(cls):
        # type: () -> Type[SetObject]
        """Base auxiliary object type."""
        return SetObject

    @property
    @final
    def _base_auxiliary_data_type(cls):
        # type: () -> Type[SetData]
        """Base auxiliary data type."""
        return SetData


# noinspection PyTypeChecker
_SO = TypeVar("_SO", bound="SetObject")


class SetObject(
    with_metaclass(
        SetObjectMeta,
        BaseAuxiliaryObject[T],
        BaseSetStructure[T],
    )
):
    """
    Set object.

    Inherits from:
      - :class:`objetto.bases.BaseAuxiliaryObject`
      - :class:`objetto.bases.BaseSetStructure`

    Inherited by:
      - :class:`objetto.objects.MutableSetObject`

    :param app: Application.
    :type app: objetto.applications.Application

    :param initial: Initial values.
    :type initial: collections.abc.Iterable[collections.abc.Hashable]
    """

    __slots__ = ()
    __functions__ = SetObjectFunctions

    @classmethod
    @final
    def _from_iterable(cls, iterable):
        # type: (Iterable) -> SetState
        """
        Make set state from iterable.

        :param iterable: Iterable.
        :type iterable: collections.abc.Iterable[collections.abc.Hashable]

        :return: Set state.
        :rtype: objetto.objects.SetObject
        """
        return SetState(iterable)

    def __init__(
        self,
        app,  # type: Application
        initial=(),  # type: Iterable[Hashable]
    ):
        # type: (...) -> None
        super(SetObject, self).__init__(app=app)
        self.__functions__.update(self, initial)

    @final
    def _clear(self):
        # type: (_SO) -> _SO
        """
        Clear all values.

        :return: Transformed.
        :rtype: objetto.objects.SetObject
        """
        self.__functions__.remove(self, self._state)
        return self

    @final
    def _add(self, value):
        # type: (_SO, T) -> _SO
        """
        Add value.

        :param value: Value.
        :type value: collections.abc.Hashable

        :return: Transformed.
        :rtype: objetto.objects.SetObject
        """
        self.__functions__.update(self, {value})
        return self

    @final
    def _discard(self, *values):
        # type: (_SO, T) -> _SO
        """
        Discard value(s).

        :param values: Value(s).
        :type values: collections.abc.Hashable

        :return: Transformed.
        :rtype: objetto.objects.SetObject

        :raises ValueError: No values provided.
        """
        with self.app.write_context():
            removes = self._state.intersection(values)
            if not removes:
                return self
            self._remove(*removes)
        return self

    @final
    def _remove(self, *values):
        # type: (_SO, T) -> _SO
        """
        Remove existing value(s).

        :param values: Value(s).
        :type values: collections.abc.Hashable

        :return: Transformed.
        :rtype: objetto.objects.SetObject

        :raises ValueError: No values provided.
        :raises KeyError: Value is not present.
        """
        self.__functions__.remove(self, values)
        return self

    @final
    def _replace(self, old_value, new_value):
        # type: (_SO, T, T) -> _SO
        """
        Replace existing value with a new one.

        :param old_value: Existing value.
        :type old_value: collections.abc.Hashable

        :param new_value: New value.
        :type new_value: collections.abc.Hashable

        :return: Transformed.
        :rtype: objetto.objects.SetObject

        :raises KeyError: Value is not present.
        """
        metadata = dict(old_value=old_value, new_value=new_value)
        with self._batch_context("Replace Value", **metadata):
            self._remove(old_value)
            self._add(new_value)
        return self

    @final
    def _update(self, iterable):
        # type: (_SO, Iterable[T]) -> _SO
        """
        Update with iterable.

        :param iterable: Iterable.
        :type iterable: collections.abc.Iterable[collections.abc.Hashable]

        :return: Transformed.
        :rtype: objetto.objects.SetObject
        """
        self.__functions__.update(self, iterable)
        return self

    @final
    def _locate(self, child):
        # type: (BaseObject) -> BaseObject
        """
        Locate child object.

        :param child: Child object.
        :type child: objetto.bases.BaseObject

        :return: Location (the child object itself).
        :rtype: objetto.bases.BaseObject

        :raises ValueError: Could not locate child.
        """
        with self.app.__.read_context(self) as read:
            if child not in read().children:
                error = "could not locate child {} in {}".format(child, self)
                raise ValueError(error)
            return child

    @final
    def _locate_data(self, child):
        # type: (BaseObject) -> BaseData
        """
        Locate child object's data.

        :param child: Child object.
        :type child: objetto.bases.BaseObject

        :return: Data location (the child data itself).
        :rtype: objetto.bases.BaseData

        :raises ValueError: Could not locate child's data.
        """
        with self.app.__.read_context(self) as read:
            if child not in read().children:
                error = "could not locate child {} in {}".format(child, self)
                raise ValueError(error)
            metadata = read().metadata
            try:
                return metadata[DATA_MAP_METADATA_KEY][child]
            except KeyError:
                pass
            error = "could not locate data of child {} in {}".format(child, self)
            raise ValueError(error)

    @classmethod
    @final
    def deserialize(cls, serialized, app=None, **kwargs):
        # type: (Type[_SO], Any, Application, Any) -> _SO
        """
        Deserialize.

        :param serialized: Serialized.
        :type serialized: list

        :param app: Application (required).
        :type app: objetto.applications.Application

        :param kwargs: Keyword arguments to be passed to the deserializers.

        :return: Deserialized.
        :rtype: objetto.objects.SetObject

        :raises objetto.exceptions.SerializationError: Can't deserialize.
        """
        assert isinstance(serialized, list)

        if app is None:
            error = (
                "missing required 'app' keyword argument for '{}.deserialize()' method"
            ).format(cls.__fullname__)
            raise ValueError(error)
        kwargs["app"] = app

        if not cls._relationship.serialized:
            error = "'{}' is not deserializable".format(cls.__name__)
            raise SerializationError(error)

        with app.write_context():
            self = cast("_SO", cls.__new__(cls))
            with init_context(self):
                super(SetObject, self).__init__(app)
                initial = set(
                    cls.deserialize_value(v, None, **kwargs)
                    for v in serialized
                    if cls._relationship.serialized
                )
                self.__functions__.update(self, initial)
            return self

    @final
    def serialize(self, **kwargs):
        # type: (Any) -> List[Hashable]
        """
        Serialize.

        :param kwargs: Keyword arguments to be passed to the serializers.

        :return: Serialized.
        :rtype: list

        :raises objetto.exceptions.SerializationError: Can't serialize.
        """
        with self.app.read_context():
            if not type(self)._relationship.serialized:
                error = "'{}' is not serializable".format(type(self).__fullname__)
                raise SerializationError(error)

            return list(
                self.serialize_value(v, None, **kwargs)
                for v in self._state
                if type(self)._relationship.serialized
            )

    @property
    @final
    def _state(self):
        # type: () -> SetState[T]
        """
        State.

        :rtype: objetto.states.LSettate
        """
        return cast("SetState[T]", super(BaseSetStructure, self)._state)

    @property
    @final
    def data(self):
        # type: () -> SetData[T]
        """
        Data.

        :rtype: objetto.data.SetData
        """
        return cast("SetData[T]", super(BaseSetStructure, self).data)


# noinspection PyAbstractClass
class MutableSetObject(
    BaseMutableSetStructure[T], SetObject[T], BaseMutableAuxiliaryObject[T]
):
    """
    Mutable set object.

    Inherits from:
      - :class:`objetto.bases.BaseMutableSetStructure`
      - :class:`objetto.objects.SetObject`
      - :class:`objetto.bases.BaseMutableAuxiliaryObject`
    """

    __slots__ = ()

    @final
    def pop(self):
        # type: () -> T
        """
        Pop value.

        :return: Value.
        :rtype: collections.abc.Hashable

        :raises KeyError: Empty set.
        """
        with self.app.write_context():
            state = self._state
            if not state:
                error = "empty set"
                raise KeyError(error)
            value = next(iter(state))
            self._remove(value)
        return value

    @final
    def intersection_update(self, iterable):
        # type: (Iterable[T]) -> None
        """
        Intersect.

        :param iterable: Iterable.
        :type iterable: collections.abc.Iterable
        """
        with self.app.write_context():
            difference = self.difference(iterable)
            if difference:
                self._remove(*difference)

    @final
    def symmetric_difference_update(self, iterable):
        # type: (Iterable[T]) -> None
        """
        Symmetric difference.

        :param iterable: Iterable.
        :type iterable: collections.abc.Iterable
        """
        with self.app.write_context():
            inverse_difference = self.inverse_difference(iterable)
            intersection = self.intersection(iterable)
            self._update(inverse_difference)
            self._remove(*intersection)

    @final
    def difference_update(self, iterable):
        # type: (Iterable[T]) -> None
        """
        Difference.

        :param iterable: Iterable.
        :type iterable: collections.abc.Iterable
        """
        with self.app.write_context():
            intersection = self.intersection(iterable)
            if intersection:
                self._remove(*intersection)


# noinspection PyTypeChecker
_PSO = TypeVar("_PSO", bound="ProxySetObject")


@final
class ProxySetObject(BaseProxyObject[T], BaseMutableSet[T]):
    """
    Mutable proxy set.

    Inherits from:
      - :class:`objetto.bases.BaseProxyObject`
      - :class:`objetto.bases.BaseMutableSet`
    """

    __slots__ = ()

    def pop(self):
        # type: () -> T
        """
        Pop value.

        :return: Value.
        :rtype: collections.abc.Hashable

        :raises KeyError: Empty set.
        """
        with self.app.write_context():
            state = self._state
            if not state:
                error = "empty set"
                raise KeyError(error)
            value = next(iter(state))
            self._remove(value)
        return value

    def intersection_update(self, iterable):
        # type: (Iterable[T]) -> None
        """
        Intersect.

        :param iterable: Iterable.
        :type iterable: collections.abc.Iterable
        """
        with self.app.write_context():
            difference = self.difference(iterable)
            if difference:
                self._remove(*difference)

    def symmetric_difference_update(self, iterable):
        # type: (Iterable[T]) -> None
        """
        Symmetric difference.

        :param iterable: Iterable.
        :type iterable: collections.abc.Iterable
        """
        with self.app.write_context():
            inverse_difference = self.inverse_difference(iterable)
            intersection = self.intersection(iterable)
            self._update(inverse_difference)
            self._remove(*intersection)

    def difference_update(self, iterable):
        # type: (Iterable[T]) -> None
        """
        Difference.

        :param iterable: Iterable.
        :type iterable: collections.abc.Iterable
        """
        with self.app.write_context():
            intersection = self.intersection(iterable)
            if intersection:
                self._remove(*intersection)

    @classmethod
    def _from_iterable(cls, iterable):
        # type: (Iterable) -> SetState
        """
        Make set state from iterable.

        :param iterable: Iterable.
        :type iterable: collections.abc.Iterable[collections.abc.Hashable]

        :return: Set state.
        :rtype: objetto.objects.SetObject
        """
        return SetState(iterable)

    def _hash(self):
        # type: () -> int
        """
        Get hash.

        :return: Hash.
        :rtype: int
        """
        return self._obj._hash()

    def _add(self, value):
        # type: (_PSO, T) -> _PSO
        """
        Add value.

        :param value: Value.
        :type value: collections.abc.Hashable

        :return: Transformed.
        :rtype: objetto.objects.ProxySetObject
        """
        self._obj._add(value)
        return self

    def _discard(self, *values):
        # type: (_PSO, T) -> _PSO
        """
        Discard value(s).

        :param values: Value(s).
        :type values: collections.abc.Hashable

        :return: Transformed.
        :rtype: objetto.objects.ProxySetObject

        :raises ValueError: No values provided.
        """
        self._obj._discard(*values)
        return self

    def _remove(self, *values):
        # type: (_PSO, T) -> _PSO
        """
        Remove existing value(s).

        :param values: Value(s).
        :type values: collections.abc.Hashable

        :return: Transformed.
        :rtype: objetto.objects.ProxySetObject

        :raises ValueError: No values provided.
        :raises KeyError: Value is not present.
        """
        self._obj._remove(*values)
        return self

    def _replace(self, old_value, new_value):
        # type: (_PSO, T, T) -> _PSO
        """
        Replace existing value with a new one.

        :param old_value: Existing value.
        :type old_value: collections.abc.Hashable

        :param new_value: New value.
        :type old_value: collections.abc.Hashable

        :return: Transformed.
        :rtype: objetto.objects.ProxySetObject

        :raises KeyError: Value is not present.
        """
        self._obj._replace(old_value, new_value)
        return self

    def _update(self, iterable):
        # type: (_PSO, Iterable[T]) -> _PSO
        """
        Update with iterable.

        :param iterable: Iterable.
        :type iterable: collections.abc.Iterable[collections.abc.Hashable]

        :return: Transformed.
        :rtype: objetto.objects.ProxySetObject
        """
        self._obj._update(iterable)
        return self

    def isdisjoint(self, iterable):
        # type: (Iterable) -> bool
        """
        Get whether is a disjoint set of an iterable.

        :param iterable: Iterable.
        :type iterable: collections.abc.Iterable

        :return: True if is disjoint.
        """
        return self._obj.isdisjoint(iterable)

    def issubset(self, iterable):
        # type: (Iterable) -> bool
        """
        Get whether is a subset of an iterable.

        :param iterable: Iterable.
        :type iterable: collections.abc.Iterable

        :return: True if is subset.
        """
        return self._obj.issubset(iterable)

    def issuperset(self, iterable):
        # type: (Iterable) -> bool
        """
        Get whether is a superset of an iterable.

        :param iterable: Iterable.
        :type iterable: collections.abc.Iterable

        :return: True if is superset.
        """
        return self._obj.issuperset(iterable)

    def intersection(self, iterable):
        # type: (Iterable) -> SetState
        """
        Get intersection.

        :param iterable: Iterable.
        :type iterable: collections.abc.Iterable

        :return: Intersection.
        """
        return self._obj.intersection(iterable)

    def difference(self, iterable):
        # type: (Iterable) -> SetState
        """
        Get difference.

        :param iterable: Iterable.
        :type iterable: collections.abc.Iterable

        :return: Difference.
        """
        return self._obj.difference(iterable)

    def inverse_difference(self, iterable):
        # type: (Iterable) -> SetState
        """
        Get an iterable's difference to this.

        :param iterable: Iterable.
        :type iterable: collections.abc.Iterable

        :return: Inverse Difference.
        """
        return self._obj.inverse_difference(iterable)

    def symmetric_difference(self, iterable):
        # type: (Iterable) -> SetState
        """
        Get symmetric difference.

        :param iterable: Iterable.
        :type iterable: collections.abc.Iterable

        :return: Symmetric difference.
        """
        return self._obj.symmetric_difference(iterable)

    def union(self, iterable):
        # type: (Iterable) -> SetState
        """
        Get union.

        :param iterable: Iterable.
        :type iterable: collections.abc.Iterable

        :return: Union.
        """
        return self._obj.union(iterable)

    @property
    def _obj(self):
        # type: () -> SetObject[T]
        """
        Set object.

        :rtype: objetto.objects.SetObject
        """
        return cast("SetObject[T]", super(ProxySetObject, self)._obj)

    @property
    def _state(self):
        # type: () -> SetState[T]
        """
        State.

        :rtype: objetto.states.SetState
        """
        return cast("SetState[T]", super(ProxySetObject, self)._state)

    @property
    def data(self):
        # type: () -> Optional[SetData[T]]
        """
        Data.

        :rtype: objetto.data.SetData or None
        """
        return cast("Optional[SetData[T]]", super(ProxySetObject, self).data)



# -*- coding: utf-8 -*-

import copy
import pickle

try:
    import collections.abc as collections_abc
except ImportError:
    import collections as collections_abc  # type: ignore

import pytest
import six
import slotted

from objetto._bases import (
    FINAL_CLASS_TAG,
    FINAL_METHOD_TAG,
    INITIALIZING_TAG,
    Base,
    BaseCollection,
    BaseContainer,
    BaseDict,
    BaseHashable,
    BaseInteractiveCollection,
    BaseInteractiveDict,
    BaseInteractiveList,
    BaseInteractiveSet,
    BaseIterable,
    BaseList,
    BaseMeta,
    BaseMutableCollection,
    BaseMutableDict,
    BaseMutableList,
    BaseMutableSet,
    BaseProtectedCollection,
    BaseProtectedDict,
    BaseProtectedList,
    BaseProtectedSet,
    BaseSet,
    BaseSized,
    abstract_member,
    final,
    init,
    init_context,
    make_base_cls,
    simplify_member_names,
)


def test_initializing():
    class MyBase(Base):
        def __init__(self):
            assert getattr(self, INITIALIZING_TAG) is True
            self._method()

        def _method(self):
            assert getattr(self, INITIALIZING_TAG) is True

        def method(self):
            assert getattr(self, INITIALIZING_TAG) is False

        @init
        def init_method(self):
            assert getattr(self, INITIALIZING_TAG) is True

        def init_method_context(self):
            with init_context(self):
                assert getattr(self, INITIALIZING_TAG) is True
            assert getattr(self, INITIALIZING_TAG) is False

    class MySubBase(MyBase):
        def __init__(self):
            super(MySubBase, self).__init__()
            assert getattr(self, INITIALIZING_TAG) is True
            self._method()

        def init_method(self):
            assert getattr(self, INITIALIZING_TAG) is False

    my_obj = MyBase()
    my_obj.method()
    my_obj.init_method()
    my_obj.init_method_context()
    assert getattr(my_obj, INITIALIZING_TAG) is False

    my_sub_obj = MySubBase()
    my_sub_obj.method()
    my_sub_obj.init_method()
    my_sub_obj.init_method_context()
    assert getattr(my_sub_obj, INITIALIZING_TAG) is False


def test_force_hash_declaration():
    with pytest.raises(TypeError):

        class MyBase(Base):
            def __eq__(self, other):
                return False

        raise AssertionError(MyBase)


def test_final_decorator():
    class MyBase(Base):
        @final
        def my_final_method(self):
            return

    assert hasattr(MyBase.__dict__["my_final_method"], FINAL_METHOD_TAG)
    assert getattr(MyBase.__dict__["my_final_method"], FINAL_METHOD_TAG) is True


def test_final_method():
    class MyBase(Base):
        @final
        def my_final_method(self):
            return

    with pytest.raises(TypeError):
        # noinspection PyFinal
        class MySubBase(MyBase):
            def my_final_method(self):
                return

        raise AssertionError(MySubBase)


def test_final_classmethod():
    class MyBase(Base):
        @classmethod
        @final
        def my_final_method(cls):
            return

    with pytest.raises(TypeError):
        # noinspection PyFinal
        class MySubBase(MyBase):
            @classmethod
            def my_final_method(cls):
                return

        raise AssertionError(MySubBase)


def test_final_staticmethod():
    class MyBase(Base):
        @staticmethod
        @final
        def my_final_method():
            return

    with pytest.raises(TypeError):
        # noinspection PyFinal
        class MySubBase(MyBase):
            @staticmethod
            def my_final_method():
                return

        raise AssertionError(MySubBase)


def test_final_property():
    class MyBase(Base):
        @property
        @final
        def my_final_property(self):
            return

    with pytest.raises(TypeError):
        # noinspection PyFinal
        class MySubBase(MyBase):
            @property
            def my_final_property(self):
                return

        raise AssertionError(MySubBase)


def test_final_class():
    @final
    class MyBase(Base):
        pass

    with pytest.raises(TypeError):
        # noinspection PyFinal
        class MySubBase(MyBase):  # type: ignore
            pass

        raise AssertionError(MySubBase)

    class MyOtherSubBase(type(Base)("MyOtherBase", (Base,), {FINAL_CLASS_TAG: 1})):
        pass

    assert MyOtherSubBase


def test_ne():
    class MyBase(Base):
        __slots__ = ("value",)

        def __init__(self, value):
            self.value = value

        __hash__ = None

        def __eq__(self, other):
            return self.value is other.value

    assert MyBase(True) == MyBase(True)
    assert MyBase(True) != MyBase(False)

    class MyOtherBase(MyBase):
        __hash__ = None

        def __eq__(self, other):
            return NotImplemented

    assert MyOtherBase(True) == MyBase(True)
    assert MyOtherBase(True) != MyBase(False)
    assert MyOtherBase(True).__ne__(MyBase(True)) is NotImplemented


def test_copy():
    base = Base()

    with pytest.raises(RuntimeError):
        _ = copy.copy(base)


class GeneratedBaseParent(object):
    GeneratedBase = None


def test_make_base_cls():
    default_base_cls = make_base_cls()
    assert default_base_cls.__name__ == Base.__name__
    assert default_base_cls.__qualname__ == Base.__fullname__
    assert default_base_cls.__module__ == Base.__module__

    base_cls = make_base_cls(
        Base, "GeneratedBaseParent.GeneratedBase", __name__, {"class_var": 10}
    )
    assert issubclass(base_cls, Base)
    assert base_cls.__name__ == "GeneratedBase"
    assert base_cls.__qualname__ == "GeneratedBaseParent.GeneratedBase"
    assert base_cls.__module__ == __name__
    assert hasattr(base_cls, "__reduce__")

    GeneratedBaseParent.GeneratedBase = base_cls

    instance = base_cls()
    assert isinstance(instance, base_cls)
    assert type(instance) is base_cls

    pickled_instance = pickle.loads(pickle.dumps(instance))
    assert pickled_instance is not instance
    assert isinstance(pickled_instance, base_cls)
    assert type(pickled_instance) is base_cls


MyGeneratedBase = make_base_cls(Base, "MyGeneratedBase", __name__, {})


class MySubBase(MyGeneratedBase):
    pass


def test_generated_base_subclass():
    assert MySubBase.__name__ != MyGeneratedBase.__name__
    try:
        my_sub_base_qual_name = MySubBase.__qualname__
    except AttributeError:  # for python 2.7
        assert MyGeneratedBase.__qualname__ == MyGeneratedBase.__name__
        assert not hasattr(object, "__qualname__")
        assert not hasattr(MySubBase, "__qualname__")
    else:
        assert my_sub_base_qual_name != MyGeneratedBase.__qualname__


def test_generated_base_subclass_instance():
    instance = MySubBase()
    assert isinstance(instance, MySubBase)
    assert type(instance) is MySubBase
    pickled_instance = pickle.loads(pickle.dumps(instance))
    assert pickled_instance is not instance
    assert isinstance(pickled_instance, MyGeneratedBase)
    assert isinstance(pickled_instance, MySubBase)
    assert type(pickled_instance) is MySubBase


def test_dir():
    def get_var():
        return None

    def get_method():
        return lambda *args: None

    def get_class_method():
        return classmethod(get_method())

    def get_static_method():
        return staticmethod(get_method())

    def get_property_method():
        return property(get_method())

    class MyBaseMeta(BaseMeta):
        (
            meta_var,
            _meta_var,
            __meta_var,
            __meta_var__,
        ) = (get_var(),) * 4
        meta_method, _meta_method, __meta_method, __meta_method__ = (get_method(),) * 4
        (
            meta_class_method,
            _meta_class_method,
            __meta_class_method,
            __meta_class_method__,
        ) = (get_class_method(),) * 4
        (
            meta_static_method,
            _meta_static_method,
            __meta_static_method,
            __meta_static_method__,
        ) = (get_static_method(),) * 4
        (
            meta_property_method,
            _meta_property_method,
            __meta_property_method,
            __meta_property_method__,
        ) = (get_property_method(),) * 4

    class MyBase(six.with_metaclass(MyBaseMeta, Base)):
        __slots__ = ("var", "_var", "__var", "__var__")

        (
            class_var,
            _class_var,
            __class_var,
            __class_var__,
        ) = (get_var(),) * 4
        method, _method, __method, __method__ = (get_method(),) * 4
        class_method, _class_method, __class_method, __class_method__ = (
            get_class_method(),
        ) * 4
        static_method, _static_method, __static_method, __static_method__ = (
            get_static_method(),
        ) * 4
        property_method, _property_method, __property_method, __property_method__ = (
            get_property_method(),
        ) * 4

        def __init__(self):
            self.var = None
            self._var = None
            self.__var = None
            self.__var__ = None

    class MySubBase(MyBase):
        pass

    for obj in (MyBase, MyBase(), MySubBase, MySubBase()):
        assert dir(obj) == list(simplify_member_names(dir(obj)))
        for member_name in dir(obj):
            getattr(obj, member_name)


def test_abstract_member():
    class AbstractClass(Base):
        some_attribute = abstract_member()

    with pytest.raises(TypeError):
        AbstractClass()

    class ConcreteClass(AbstractClass):
        some_attribute = (1, 2, 3)

    obj = ConcreteClass()
    assert obj


def test_required_overrides():
    def check_base(base, slotted_base):
        assert issubclass(slotted_base, slotted.SlottedABC)
        exclude = {
            "__weakref__",
            "__metaclass__",
            "__setstate__",
            "__getstate__",
            "__slots__",
            "_MutableMapping__marker",
            "_abc_negative_cache_version",
        }
        assert issubclass(base, Base)
        assert issubclass(base, slotted_base)
        for open_attribute in getattr(base, "_BaseMeta__open_attributes")[base]:
            if open_attribute in exclude:
                continue
            member = getattr(base, open_attribute)
            try:
                original = getattr(slotted_base, open_attribute)
            except AttributeError:
                continue
            object_original = getattr(object, open_attribute, None)
            if member is None and original is None:
                continue
            if member is original and member is not object_original:
                raise AssertionError(
                    "'{}' does not override '{}.{}'".format(
                        base.__name__, slotted_base.__name__, open_attribute
                    )
                )

    bases = (
        (BaseHashable, slotted.SlottedHashable),
        (BaseSized, slotted.SlottedSized),
        (BaseIterable, slotted.SlottedIterable),
        (BaseContainer, slotted.SlottedContainer),
        (BaseDict, slotted.SlottedMapping),
        (BaseMutableDict, slotted.SlottedMutableMapping),
        (BaseList, slotted.SlottedSequence),
        (BaseMutableList, slotted.SlottedMutableSequence),
        (BaseSet, slotted.SlottedSet),
        (BaseMutableSet, slotted.SlottedMutableSet),
    )

    for b, s in bases:
        check_base(b, s)


def test_inheritance():

    assert issubclass(BaseMeta, slotted.SlottedABCMeta)
    assert issubclass(Base, slotted.SlottedABC)
    assert isinstance(Base, BaseMeta)

    assert issubclass(BaseHashable, Base)
    assert issubclass(BaseHashable, slotted.SlottedHashable)

    assert issubclass(BaseSized, Base)
    assert issubclass(BaseSized, slotted.SlottedSized)

    assert issubclass(BaseIterable, Base)
    assert issubclass(BaseIterable, slotted.SlottedIterable)

    assert issubclass(BaseContainer, Base)
    assert issubclass(BaseContainer, slotted.SlottedContainer)

    assert issubclass(BaseCollection, BaseContainer)
    assert issubclass(BaseCollection, BaseSized)
    assert issubclass(BaseCollection, BaseIterable)

    assert issubclass(BaseProtectedCollection, BaseCollection)
    assert issubclass(BaseInteractiveCollection, BaseProtectedCollection)
    assert issubclass(BaseMutableCollection, BaseProtectedCollection)
    assert not issubclass(BaseMutableCollection, BaseInteractiveCollection)
    assert not issubclass(BaseInteractiveCollection, BaseMutableCollection)

    assert issubclass(BaseDict, BaseCollection)
    assert not issubclass(BaseDict, collections_abc.Hashable)
    assert issubclass(BaseDict, slotted.SlottedMapping)
    assert issubclass(BaseProtectedDict, BaseProtectedCollection)
    assert issubclass(BaseInteractiveDict, BaseInteractiveCollection)
    assert issubclass(BaseMutableDict, BaseMutableCollection)
    assert issubclass(BaseMutableDict, slotted.SlottedMutableMapping)

    assert issubclass(BaseProtectedDict, BaseDict)
    assert issubclass(BaseInteractiveDict, BaseProtectedDict)
    assert issubclass(BaseMutableDict, BaseProtectedDict)
    assert not issubclass(BaseMutableDict, BaseInteractiveDict)
    assert not issubclass(BaseInteractiveDict, BaseMutableDict)

    assert issubclass(BaseList, BaseCollection)
    assert not issubclass(BaseList, collections_abc.Hashable)
    assert issubclass(BaseList, slotted.SlottedSequence)
    assert issubclass(BaseProtectedList, BaseProtectedCollection)
    assert issubclass(BaseInteractiveList, BaseInteractiveCollection)
    assert issubclass(BaseMutableList, BaseMutableCollection)
    assert issubclass(BaseMutableList, slotted.SlottedMutableSequence)

    assert issubclass(BaseProtectedList, BaseList)
    assert issubclass(BaseInteractiveList, BaseProtectedList)
    assert issubclass(BaseMutableList, BaseProtectedList)
    assert not issubclass(BaseMutableList, BaseInteractiveList)
    assert not issubclass(BaseInteractiveList, BaseMutableList)

    assert issubclass(BaseSet, BaseCollection)
    assert not issubclass(BaseSet, collections_abc.Hashable)
    assert issubclass(BaseSet, slotted.SlottedSet)
    assert issubclass(BaseProtectedSet, BaseProtectedCollection)
    assert issubclass(BaseInteractiveSet, BaseInteractiveCollection)
    assert issubclass(BaseMutableSet, BaseMutableCollection)
    assert issubclass(BaseMutableSet, slotted.SlottedMutableSet)

    assert issubclass(BaseProtectedSet, BaseSet)
    assert issubclass(BaseInteractiveSet, BaseProtectedSet)
    assert issubclass(BaseMutableSet, BaseProtectedSet)
    assert not issubclass(BaseMutableSet, BaseInteractiveSet)
    assert not issubclass(BaseInteractiveSet, BaseMutableSet)


if __name__ == "__main__":
    pytest.main()
